Package: sftrack
Title: Modern Classes for Tracking and Movement Data
Version: 0.5.5
Date: 2026-05-17
Authors@R: c(
    person(given = "Matthew",
           family = "Boone",
           role = "aut",
           email = "mttboone@gmail.com",
           comment = c(ORCID = "0000-0002-8070-4715")),
    person(given = "Rocío",
           family = "Joo",
           role = "aut",
           email = "rocio.joo@globalfishingwatch.org",
           comment = c(ORCID = "0000-0003-0319-4210")),
    person(given = "Mathieu",
           family = "Basille",
           role = c("aut", "cre"),
           email = "mathieu@basille.org",
           comment = c(ORCID = "0000-0001-9366-7127")))
Depends: R (>= 3.2.0)
Imports: sf
Description: Modern classes for tracking and movement data, building
    on 'sf' spatial infrastructure, and early theoretical work from
    Turchin (1998, ISBN: 9780878938476), and Calenge et al. (2009)
    <doi:10.1016/j.ecoinf.2008.10.002>. Tracking data are series of
    locations with at least 2-dimensional spatial coordinates (x,y), a
    time index (t), and individual identification (id) of the object
    being monitored; movement data are made of trajectories, i.e. the
    line representation of the path, composed by steps (the
    straight-line segments connecting successive locations). 'sftrack'
    is designed to handle movement of both living organisms and
    inanimate objects.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
VignetteBuilder: knitr
URL: https://mablab.org/sftrack/, https://github.com/mablab/sftrack
BugReports: https://github.com/mablab/sftrack/issues
Suggests: testthat (>= 2.1.0), utils, ggplot2, adehabitatLT, knitr,
        geosphere, scales, covr, rmarkdown, lwgeom
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-18 07:00:39 UTC; mathieu
Author: Matthew Boone [aut] (ORCID: <https://orcid.org/0000-0002-8070-4715>),
  Rocío Joo [aut] (ORCID: <https://orcid.org/0000-0003-0319-4210>),
  Mathieu Basille [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9366-7127>)
Maintainer: Mathieu Basille <mathieu@basille.org>
Repository: CRAN
Date/Publication: 2026-05-18 07:30:02 UTC
Built: R 4.4.3; ; 2026-05-18 08:05:44 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$<-",s_group)
S3method("[",c_grouping)
S3method("[",sftrack)
S3method("[",sftraj)
S3method("[<-",c_grouping)
S3method("[<-",s_group)
S3method("[[<-",sftrack)
S3method("[[<-",sftraj)
S3method("active_group<-",c_grouping)
S3method("active_group<-",sftrack)
S3method("active_group<-",sftraj)
S3method(as.data.frame,c_grouping)
S3method(as_sftrack,data.frame)
S3method(as_sftrack,ltraj)
S3method(as_sftrack,sf)
S3method(as_sftrack,sftraj)
S3method(as_sftraj,data.frame)
S3method(as_sftraj,ltraj)
S3method(as_sftraj,sf)
S3method(as_sftraj,sftrack)
S3method(c,c_grouping)
S3method(c,s_group)
S3method(cbind,sftrack)
S3method(format,c_grouping)
S3method(format,s_group)
S3method(geom_sftrack,sftrack)
S3method(geom_sftrack,sftraj)
S3method(group_labels,c_grouping)
S3method(group_labels,sftrack)
S3method(group_labels,sftraj)
S3method(group_names,c_grouping)
S3method(group_names,sftrack)
S3method(group_names,sftraj)
S3method(plot,sftrack)
S3method(plot,sftraj)
S3method(print,c_grouping)
S3method(print,s_group)
S3method(print,sftrack)
S3method(print,sftraj)
S3method(rbind,sftrack)
S3method(rbind,sftraj)
S3method(st_coordinates,sftraj)
S3method(str,c_grouping)
S3method(summary,c_grouping)
S3method(summary,sftrack)
S3method(summary,sftraj)
export("active_group<-")
export(active_group)
export(as_sftrack)
export(as_sftraj)
export(calc_sort_index)
export(check_NA_coords)
export(check_NA_group)
export(check_group_id)
export(check_group_names)
export(check_names_exist)
export(check_ordered)
export(check_t_regular)
export(check_time)
export(coord_traj)
export(dup_timestamp)
export(fix_zero)
export(geom_sftrack)
export(group_labels)
export(group_names)
export(is_linestring)
export(make_c_grouping)
export(make_s_group)
export(make_step_geom)
export(merge_traj)
export(new_sftrack)
export(new_sftraj)
export(pts_traj)
export(step_metrics)
export(step_recalc)
export(summary_sftrack)
export(which_duplicated)
import(sf)
importFrom(graphics,lcm)
importFrom(graphics,plot)
importFrom(stats,aggregate)
importFrom(utils,str)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# sftrack: Modern classes for tracking and movement data <img src="man/figures/sftrack-logo-200-transp-bg.png" align="right" width="200px"/>

<!-- badges: start -->

[![Project Status: Active - The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![Licence](https://img.shields.io/badge/license-MIT-green)](https://opensource.org/license/mit)
[![CRAN
Status](http://www.r-pkg.org/badges/version/sftrack)](https://CRAN.R-project.org/package=sftrack)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/sftrack)](https://CRAN.R-project.org/package=sftrack)
<!-- [![Build Status: master](https://travis-ci.org/mablab/sftrack.svg?branch=master)](https://travis-ci.org/mablab/sftrack)  -->
<!-- [![Build Status: -->
<!-- dev](https://img.shields.io/travis/mablab/sftrack/dev.svg?label=dev+build)](https://travis-ci.org/mablab/sftrack)  -->
[![R-CMD-check](https://github.com/mablab/sftrack/workflows/R-CMD-check/badge.svg)](https://github.com/mablab/sftrack/actions)
<!-- badges: end -->

`sftrack` provides modern classes for **tracking and movement data**,
relying on `sf` spatial infrastructure. Tracking data are made of
tracks, i.e. series of locations with at least 2-dimensional spatial
coordinates (x,y), a time index (t), and individual identification (id)
of the object being monitored; movement data are made of trajectories,
i.e. the line representation of the path, composed by steps (the
straight-line segments connecting successive locations). `sftrack` is
designed to handle movement of both living organisms and inanimate
objects.

<figure>
<img src="man/figures/definitions-512.png" title="data definitions"
alt="data definitions" />
<figcaption aria-hidden="true">data definitions</figcaption>
</figure>

The development and design of the `sftrack` package follow three simple
principles:

1.  **Minimal and focused:** this is basically the Unix philosophy. Do a
    simple thing, and do it well. The scope of the package is limited
    (see above), with as few dependencies as possible;
2.  **User-friendly:** `sftrack` is designed to be as easy to use as
    familiar R structures like `data.frame`s and `sf` objects. `sftrack`
    objects are tidy, and follow the idea that rows are records
    (locations) and columns are variable (following the semantics of
    tracking and movement data);
3.  **Flexible and extensible:** `sftrack` is meant first for users to
    use on their data, but also directly designed to address other
    developers’ needs for their own tracking packages.

## Getting started

To get started, install `sftrack` directly from CRAN, or check the
development version on GitHub with the
[remotes](https://cran.r-project.org/package=remotes) package:

``` r
# To install the stable version from CRAN
install.packages("sftrack")

# To install the dev version with built vignettes
remotes::install_github("mablab/sftrack", ref = "dev", build_vignettes = TRUE)
```

The `dev` version is updated much more frequently and should pass the
majority of CRAN checks. However, if you install the `dev` version,
understand it may still contain some bugs. Please submit any bug you
find to the [issues](https://github.com/mablab/sftrack/issues) page.

## A minimal introduction to `sftrack` and `sftraj` objects

The easiest way to create an `sftrack` object is to start from a
`data.frame` with all information as columns, typically the raw data
extracted from telemetry devices:

``` r
library("sftrack")
data(raccoon)
raccoon$timestamp <- as.POSIXct(as.POSIXlt(raccoon$timestamp, tz = "EST5EDT"))
head(raccoon)
#>   animal_id latitude longitude           timestamp height hdop vdop fix
#> 1   TTP-058       NA        NA 2019-01-18 19:02:30     NA  0.0  0.0  NO
#> 2   TTP-058 26.06945 -80.27906 2019-01-18 20:02:30      7  6.2  3.2  2D
#> 3   TTP-058       NA        NA 2019-01-18 21:02:30     NA  0.0  0.0  NO
#> 4   TTP-058       NA        NA 2019-01-18 22:02:30     NA  0.0  0.0  NO
#> 5   TTP-058 26.06769 -80.27431 2019-01-18 23:02:30    858  5.1  3.2  2D
#> 6   TTP-058 26.06867 -80.27930 2019-01-19 00:02:30    350  1.9  3.2  3D
```

In order to convert your raw data into an `sftrack` object, use the
function `as_sftrack()`. The function requires the three main elements
of tracking data:

- coordinates of the locations in at least the `x` and `y` axes (can be
  UTM, lat/long, etc., with projection provided in `crs`);
- timestamps of the locations as `POSIXct` (*or* as `integer`);
- grouping information (referred to as a “groups”), providing at least
  the identity of each individual.

``` r
my_sftrack <- as_sftrack(
  data = raccoon,
  coords = c("longitude","latitude"),
  time = "timestamp",
  group = "animal_id",
  crs = 4326)
head(my_sftrack)
#> Sftrack with 6 features and 10 fields (3 empty geometries) 
#> Geometry : "geometry" (XY, crs: WGS 84) 
#> Timestamp : "timestamp" (POSIXct in EST5EDT) 
#> Groupings : "sft_group" (*id*) 
#> -------------------------------
#>   animal_id latitude longitude           timestamp height hdop vdop fix
#> 1   TTP-058       NA        NA 2019-01-18 19:02:30     NA  0.0  0.0  NO
#> 2   TTP-058 26.06945 -80.27906 2019-01-18 20:02:30      7  6.2  3.2  2D
#> 3   TTP-058       NA        NA 2019-01-18 21:02:30     NA  0.0  0.0  NO
#> 4   TTP-058       NA        NA 2019-01-18 22:02:30     NA  0.0  0.0  NO
#> 5   TTP-058 26.06769 -80.27431 2019-01-18 23:02:30    858  5.1  3.2  2D
#> 6   TTP-058 26.06867 -80.27930 2019-01-19 00:02:30    350  1.9  3.2  3D
#>       sft_group                   geometry
#> 1 (id: TTP-058)                POINT EMPTY
#> 2 (id: TTP-058) POINT (-80.27906 26.06945)
#> 3 (id: TTP-058)                POINT EMPTY
#> 4 (id: TTP-058)                POINT EMPTY
#> 5 (id: TTP-058) POINT (-80.27431 26.06769)
#> 6 (id: TTP-058)  POINT (-80.2793 26.06867)
```

``` r
summary_sftrack(my_sftrack)
#>     group points NAs          begin_time            end_time length_m
#> 1 TTP-041    223   0 2019-01-18 19:02:30 2019-02-01 18:02:07 10212.55
#> 2 TTP-058    222   0 2019-01-18 19:02:30 2019-02-01 18:02:30 26893.27
```

While `sftrack` objects contain tracking data (locations), they can
easily be converted to movement data (with a step model instead) with
`as_sftraj`:

``` r
my_sftraj <- as_sftraj(my_sftrack)
head(my_sftraj)
#> Sftraj with 6 features and 10 fields (3 empty geometries) 
#> Geometry : "geometry" (XY, crs: WGS 84) 
#> Timestamp : "timestamp" (POSIXct in EST5EDT) 
#> Grouping : "sft_group" (*id*) 
#> -------------------------------
#>   animal_id latitude longitude           timestamp height hdop vdop fix
#> 1   TTP-058       NA        NA 2019-01-18 19:02:30     NA  0.0  0.0  NO
#> 2   TTP-058 26.06945 -80.27906 2019-01-18 20:02:30      7  6.2  3.2  2D
#> 3   TTP-058       NA        NA 2019-01-18 21:02:30     NA  0.0  0.0  NO
#> 4   TTP-058       NA        NA 2019-01-18 22:02:30     NA  0.0  0.0  NO
#> 5   TTP-058 26.06769 -80.27431 2019-01-18 23:02:30    858  5.1  3.2  2D
#> 6   TTP-058 26.06867 -80.27930 2019-01-19 00:02:30    350  1.9  3.2  3D
#>       sft_group                       geometry
#> 1 (id: TTP-058)                    POINT EMPTY
#> 2 (id: TTP-058)     POINT (-80.27906 26.06945)
#> 3 (id: TTP-058)                    POINT EMPTY
#> 4 (id: TTP-058)                    POINT EMPTY
#> 5 (id: TTP-058) LINESTRING (-80.27431 26.06...
#> 6 (id: TTP-058) LINESTRING (-80.2793 26.068...
```

Both objects can easily be plotted with base R plot functions, which
highlights the fundamental difference between tracking and movement data
(`sftrack` on the left; `sftraj` on the right):

``` r
plot(my_sftrack, main = "Tracking data (locations)")
plot(my_sftraj, main = "Movement data (steps)")
```

<img src="man/figures/README-plot-1.png" alt="" width="50%" /><img src="man/figures/README-plot-2.png" alt="" width="50%" />

## Roadmap

- Data class converters from the main tracking packages, such as
  `move::Move` and `trackeR::trackeRdata`, integrated into `as_sftrack`;
- More plotting options for tracks and trajectories (in base R and
  `ggplot2`);
- Provide Gantt chart-like or chronogram-like graphs;
- Dynamic exploration of trajectories.

## How can you help?

1.  **Submit any bug you find to the
    [issues](https://github.com/mablab/sftrack/issues) page;**
2.  Address open questions (see below);
3.  Contribute use cases (see below).

### Address open questions: *We need your feedback!*

While the foundations of the package are now pretty solid, we are still
dealing with open questions about several aspects of the package,
including the names of `sftrack` variables (e.g. coordinates,
timestamps, id and error), the structure of the grouping factor, or the
structure of the error term.

If you have strong opinions or simply want to help on the technical
side, we invite you to comment on those [open issues
here](https://github.com/mablab/sftrack/labels/question).

### Contribute use cases: *We need your feedback!*

We also need to precisely understand what is expected from such a
package. The idea here is to collect all possible use cases for a
trajectory object in R. We know they are multiple, and will contribute
our own use cases — however, we want `sftrack` to be as useful as
possible, and to act as a center piece for movement in R, so we need you
to tell us how you would use it. **In other words, we want to understand
what you expect from such a package, as a user or as a developer.** For
this, we ask you to fill out special issues in the GitHub tracker of the
package, following the [‘Use case’
template](https://github.com/mablab/sftrack/issues/new?assignees=&labels=&template=use-case.md&title=%5BUse+case%5D+Change+this+title).

Use cases do not need to be very complicated, but need to present a
specific use in human terms, the technical requirements associated to
it, and the input and output of the use case. Such use case could look
like this:

> **\[Use case\] Amazing plot for trajectory**
>
> **Use case:**
>
> Plot a trajectory using my `special_trajplot` function, which shows
> \[something amazing\].
>
> **Requirements:**
>
> - spatial coordinates (x,y) as geographic coordinates with projection
>   information
>
> - a time (t) as POSIXt object, ordered in time
>
> - information that identifies individuals (e.g. animal) for each
>   location
>
> - data associated to each location directly accessible
>
> **Input:** a `sftrack` object
>
> **Output:** a plot with \[something amazing\] about the trajectory
>
> **Additional information:** See my `special_trajplot` function here
> \[with link\].

Another example could be like this:

> **\[Use case\] Fill in missing locations in a sequence**
>
> **Use case:** Fill in the missing locations of a trajectory that
> contains spatial or temporal gaps. (for instance coming from GPS with
> failed fixes); In other words add in the missing values of a
> trajectory, i.e. timestamps with no geographic coordinates.
>
> **Requirements:**
>
> - a time (t) as POSIXt object, ordered in time
>
> - information that identifies sequences of locations (optional, if
>   several sequences), which could be different circuits of one
>   individual, or different individuals, etc.
>
> - `sftrack` should be capable of handling/storing missing values
>
> **Input:** a `sftrack` object
>
> **Output:** a `sftrack` object with additional timestamps for gaps
> (but otherwise identical in every way to the original `sftrack`)
>
> **Additional information:** See `adehabitatLT::setNA`, which does
> exactly that on `ltraj` objects.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)


## -----------------------------------------------------------------------------
library("sftrack")

#data
data("raccoon", package = "sftrack")

#xyz
coords <- raccoon[,c("longitude", "latitude")]
crs <- 4326
#groupings
group <- list(id = raccoon$animal_id,month = as.POSIXlt(raccoon$timestamp)$mon+1)
active_group <- c("id","month")
#time
time <- as.POSIXct(raccoon$timestamp, tz = "EST")
#error
error <- raccoon$fix
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, 
                         active_group = active_group, time = time, 
                         crs = crs, error = error)

head(my_sftrack)


## -----------------------------------------------------------------------------
raccoon$time <- as.POSIXct(raccoon$timestamp, tz = "EST")
raccoon$month <- as.POSIXlt(raccoon$timestamp)$mon + 1

coords <- c("longitude", "latitude")
group <- c(id = "animal_id", month = "month")
time <- "time"
error <- "fix"

my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time, error = error)

head(my_sftraj)

## ----message = FALSE----------------------------------------------------------
library("adehabitatLT")

ltraj_df <- as.ltraj(xy = raccoon[,c('longitude','latitude')], date = as.POSIXct(raccoon$timestamp),
 id = raccoon$animal_id, typeII = TRUE,
 infolocs = raccoon[,1:6])

my_sf <- as_sftrack(ltraj_df)
head(my_sf)


## -----------------------------------------------------------------------------
library("sf")
df1 <- raccoon[!is.na(raccoon$latitude),]
sf_df <- st_as_sf(df1, coords=c("longitude","latitude"), crs = crs)
group <- c(id = "animal_id")
time_col <- "time"

new_sftraj <- as_sftraj(sf_df, group = group, time = time_col) 
head(new_sftraj)

new_sftrack <- as_sftrack(sf_df, group = group, time = time_col) 
head(new_sftrack)


## -----------------------------------------------------------------------------
# Make tracks from raw data
coords <- c("longitude","latitude")
group <- c(id = "animal_id", month = "month")
time <- "time"
error <- "fix"

my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time, error = error)
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error)

# Convert between types
new_sftrack <- as_sftrack(my_sftraj)
#head(new_sftrack)
new_sftraj <- as_sftraj(my_sftrack)
#head(new_sftraj)

identical(my_sftraj,new_sftraj)
identical(my_sftrack,new_sftrack)

## -----------------------------------------------------------------------------

raccoon$time[1] <- raccoon$time[2]
try(as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error))


## -----------------------------------------------------------------------------

which_duplicated(data = raccoon , group = group, time = time)
raccoon <- raccoon[-2,]
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)


## ----out.width = "75%", fig.align = "center", fig.cap = "Fig.1: Glossary introducing the main concepts.", echo = FALSE----
knitr::include_graphics("glossary.png")


## ----sftrack-overview, echo = FALSE-------------------------------------------
library("sftrack")
data(raccoon)
raccoon$timestamp = as.POSIXct(raccoon$timestamp, tz = "EST5EDT")
my_sftrack <- as_sftrack(
  data = raccoon,
  coords = c("longitude","latitude"),
  time = "timestamp",
  group = "animal_id",
  crs = 4326)
head(my_sftrack)
plot(my_sftrack)


## ----sftraj-overview, echo = FALSE--------------------------------------------
my_sftraj <- as_sftraj(my_sftrack)
head(my_sftraj)
plot(my_sftraj)

#' @title Calculate step geometries given a set of groupings, time, and geometries
#'
#' @description This calculates step geometries as individual line segments based on the active_group
#' @param group a c_grouping object
#' @param time_data time vector
#' @param geometry the geometery data from either sf or sf_track. Must be an sf geometry class
#' @export make_step_geom
#' @examples
#' #'
#' library("sf")
#' geom <- st_as_sf(data.frame(
#'   x = c(1, 2, 2, 5),
#'   y = c(0, 1, 5, 7)
#' ), coords = c("x", "y"))
#'
#' burst <- list(id = rep(1, 4))
#' time <- 1:4
#'
#' cg <- make_c_grouping(burst)
#'
#' make_step_geom(
#'   group = cg,
#'   geometry = geom$geometry,
#'   time_data = time
#' )
make_step_geom <- function(group, time_data, geometry) {
  # burstz <- list(id = raccoon_data$animal_id, month = as.POSIXlt(raccoon_data$timestamp)$mon)
  # #data_sf <- new_sftrack(raccoon_data, time =as.POSIXct(raccoon_data$timestamp),error = NA, coords = c('longitude','latitude','height'), tz = 'UTC',burst =burstz)
  # burst = burst_select(make_multi_burst(burstz, active_burst = c('id')))
  # time_data = raccoon_data$timestamp

  # if theres more than one burst, then we combine bursts
  # if (length(burst[[1]]) > 1) {
  #   message('more than one burst selected, bursts will be combined for step geometry')
  # }

  idz <- group_labels(group)
  #
  unique_idz <- levels(idz)[table(idz) > 0]

  step_geometry <- rep(NA, length(geometry))

  # check dimensions
  point_d <- class(geometry[[1]])[1]
  nd <- which(point_d == c(NA, "XY", "XYZ"))

  for (i in unique_idz) {
    #  i <- unique_idz[1]
    subz <- idz == i
    # need to order step geometry
    order_t <- order(time_data[subz])

    sub_geom <- geometry[subz]
    sub_geom <- sub_geom[order_t]
    # We cant actually inject a null point and have it convert to line string, so we have to deal with that later

    x1 <- sub_geom[1:length(sub_geom)]
    x2 <-
      c(
        sf::st_sfc(sub_geom[2:(length(sub_geom))]),
        sf::st_sfc(st_point(rep(NA_real_, nd),
          dim = point_d
        ), crs = sf::st_crs(sub_geom))
      )
    first_point <- min(which(subz))

    x3 <- mapply(function(x, y) {
      # x <- x1[[1]]
      # y <- x2[[1]]
      if (any(c(is.na(x), is.na(y)))) {
        if (all(!is.na(x))) {
          new_geom <- sf::st_point(x)
        } else {
          new_geom <- sf::st_point()
        }
      } else {
        new_geom <- sf::st_linestring(rbind(x, y))
      }
      new_geom
    }, x1, x2, SIMPLIFY = FALSE)
    sf_x <- x3
    step_geometry[subz] <- sf_x[order(order_t)]
  }

  return(sf::st_sfc(step_geometry, crs = attr(geometry, "crs")))
}

# step function
#' @title Calculates step metrics including distance, dt, dx, and dy.
#' @param sftraj an sftrack/sftraj object. sftrack objects will be converted to sftraj internally for calculation.
#' @export
#' @examples
#' #'
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
#' # Input is a data.frame
#' my_sftraj <- as_sftraj(raccoon,
#'   group = burstz, time = "timestamp",
#'   error = NA, coords = c("longitude", "latitude")
#' )
#'
#' step_metrics(my_sftraj)[1:10, ]
step_metrics <- function(sftraj) {
  if (inherits(sftraj, "sftrack")) {
    sftraj <- as_sftraj(sftraj)
  }
  group_col <- attr(sftraj, "group_col")
  time_col <- attr(sftraj, "time_col")
  sftraj$sftrack_id <-
    paste0(group_labels(sftraj[[group_col]]), "_", sftraj[[time_col]])
  order_t <- order(group_labels(sftraj[[group_col]]), sftraj[[time_col]])
  sftraj <- sftraj[order_t, ]
  is_latlong <- any(st_is_longlat(sftraj), na.rm = TRUE)


  ret <-
    lapply(levels(group_labels(sftraj[[group_col]])), function(index) {
      # index = levels(burst_labels(sftraj$burst, factor = TRUE))[1]
      sub <- sftraj[group_labels(sftraj[[group_col]]) == index, ]
      # if only 1 row
      if (nrow(sub) == 1) {
        return(
          data.frame(
            dx = NA,
            dy = NA,
            dist = NA,
            dt = NA,
            abs_angle = NA,
            speed = NA,
            sftrack_id = sub$sftrack_id
          )
        )
      }

      dx <- get_dx(sub)
      dy <- get_dy(sub)
      dist <- as.numeric(st_length(sub))

      time <- sub[[attr(sub, "time_col")]]
      dt <- c(unclass(time[-1]) - unclass(time[-length(time)]), NA)

      if (!is_latlong) {
        abs_angle <- ifelse(dist < 1e-07, NA, atan2(dy, dx))
      } else {
        x1 <- coord_traj(sub[[attr(sub, "sf_column")]])
        x2 <- rbind(x1[-1, ], c(NA, NA))
        abs_angle <- c((geosphere::bearing(x1[-nrow(x1), ], x2[-nrow(x2), ]) - 90) * -pi / 180, NA)
        abs_angle[!is.na(abs_angle) & abs_angle > (pi)] <- abs_angle[!is.na(abs_angle) & abs_angle > (pi)] - 2 * pi
      }
      dist[is.na(dx) | is.na(dy)] <- NA

      speed <- ifelse(is.na(dist), NA, dist / dt)


      rel_angle <- c(NA, abs_angle[-1] - abs_angle[-length(abs_angle)])
      rel_angle <- ifelse(rel_angle <= (-pi), 2 * pi + rel_angle, rel_angle)
      rel_angle <- ifelse(rel_angle > pi, rel_angle - 2 * pi, rel_angle)

      so <- cbind.data.frame(
        dx = dx,
        dy = dy,
        dist = dist,
        dt = dt,
        abs_angle = abs_angle,
        rel_angle = rel_angle,
        speed = speed
      )
      so[nrow(so), c("dx", "dy", "dist", "dt", "abs_angle", "rel_angle", "speed")] <- NA
      so$sftrack_id <- sub$sftrack_id
      return(so)
    })
  ret <- do.call(rbind, ret)

  ret[order(order_t), ]
}

#' @title recalculate step geometry
#' @description Step geometeries in sftraj objects are linestrings going from t1 to t2 of a 'step'. As these are stored at the row level they are not dynamic to changes in t2.
#' step_recalc allows you to recalculate these geometeries if your data.frame has changed because of subsetting or filtering.
#' @param x an sftraj object.
#' @param return return step_geometry instead of replacing sftraj object with new step geometry. Defaults to FALSE
#' @export
step_recalc <- function(x, return = FALSE) {
  if (!inherits(x, "sftraj")) {
    stop("object is not an sftraj object")
  }
  att <- attributes(x)
  time_col <- att$time_col
  sf_col <- att$sf_column
  group_col <- att$group_col
  geom <- pts_traj(x[[sf_col]], sfc = T)
  step_geometry <-
    make_step_geom(
      group = x[[group_col]],
      geometry = geom,
      time_data = x[[time_col]]
    )
  if (return) {
    return(step_geometry)
  }

  x[[sf_col]] <- step_geometry
  x
}

get_dx <- function(x) {
  # position = 'x2'
  x <- st_geometry(x)
  crs <- st_crs(x)
  ret <- vapply(
    x, function(y) {
      # y = st_geometry(x)[[1]]
      if (inherits(y, "LINESTRING")) {
        st_length(st_sfc(st_linestring(rbind(c(y[1], y[3]), c(y[2], y[3]))), crs = crs))
      } else {
        0
      }
    },
    numeric(1)
  )
  ret[st_is_empty(x)] <- NA
  ret
}


get_dy <- function(x) {
  # position = 'x2'
  x <- st_geometry(x)
  crs <- st_crs(x)
  ret <- vapply(
    x, function(y) {
      # y = st_geometry(x)[[1]]
      if (inherits(y, "LINESTRING")) {
        st_length(st_sfc(st_linestring(rbind(c(y[1], y[3]), c(y[1], y[4]))), crs = crs))
      } else {
        0
      }
    },
    numeric(1)
  )

  ret[st_is_empty(x)] <- NA
  ret
}

test_that("step geometry calculates correctly", {
  df1 <- data.frame(
    id = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = 0:3,
    timez = Sys.time() + 60 * 60 * (1:4)
  )
  geom_ans <- c(27, 27, -80, -81, 27, -81, 27, 27, -82, -83, 27, -83)
  my_track <- as_sftrack(data = df1, group = "id", time = "timez", coords = c("x", "y"))

  my_step_geom <- make_step_geom(group = my_track$sft_group, geometry = my_track$geometry, time_data = my_track$timez)
  # order was done correctly

  expect_equal(unlist(my_step_geom), geom_ans)
  # object is an sf line string
  expect_equal(class(my_step_geom)[1], c("sfc_GEOMETRY"))

  # multiple bursts become combined
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = 0:3,
    timez = Sys.time() + 60 * 60 * (1:4)
  )
  my_track <- as_sftrack(data = df1, group = c("id", "month"), time = "timez", coords = c("x", "y"))

  my_step_geom <- suppressMessages(make_step_geom(group = my_track$sft_group, geometry = my_track$geometry, time_data = my_track$time))

  expect_equal(unlist(my_step_geom), geom_ans)

  #
})

test_that("step metrics works", {
  df1 <- data.frame(
    id = c(1, 1, 1, 1, 1, 1),
    month = c(1, 1, 1, 1, 1, 1),
    x = c(1, 1, 1, NA, 2, 3),
    y = c(0, 1, 2, NA, 3, 3),
    z = c(0, 1, 2, 3, 4, 5),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:6)
  )
  my_sftraj <- as_sftraj(
    data = df1, group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id", "month"), coords = df1[, c("x", "y")]
  )
  sm <- step_metrics(my_sftraj)
  ans <- data.frame(
    dx = c(0, 0, 0, NA, 1, NA),
    dy = c(1, 1, 0, NA, 0, NA),
    dist = c(1, 1, 0, NA, 1, NA),
    dt = c(3600, 3600, 3600, 3600, 3600, NA),
    abs_angle = c(1.570796, 1.570796, NA, NA, 0, NA),
    rel_angle = c(NA, 0, NA, NA, NA, NA)
  )
  ans$speed <- ifelse(is.na(ans$dist), NA, ans$dist / ans$dt)

  expect_equivalent(ans, sm[, 1:7], tolerance = 1e-06)

  # Test that step_metrics returns the correct order when given a numeric
  df1 <- data.frame(
    id = c(1, 1, 1, 1, 1, 1),
    month = c(1, 1, 1, 1, 1, 1),
    x = c(1, 1, 1, NA, 2, 3),
    y = c(0, 1, 2, NA, 3, 3),
    z = c(0, 1, 1, 3, 4, 5),
    timez = c(1, 2, 3, 4, 5, 6)
  )
  my_sftraj <- as_sftraj(
    data = df1, group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id", "month"), coords = df1[, c("x", "y")]
  )
  expect_equal(step_metrics(my_sftraj)$sftrack_id, c("1_1_1", "1_1_2", "1_1_3", "1_1_4", "1_1_5", "1_1_6"))

  # even if unorder manually
  my_sftraj <- my_sftraj[c(1, 3, 2, 4, 5, 6), ]
  expect_equal(step_metrics(my_sftraj)$sftrack_id, c("1_1_1", "1_1_3", "1_1_2", "1_1_4", "1_1_5", "1_1_6"))
})

#' @name grouping-class
#' @title A class to group movement data
#' @description This class describes grouping variables for movement data.
#' The grouping object is composed of a list with named vectors.
#' One of which must be 'id', this is the id of subject being monitored (commonly animal id in movement data)
#' Can be any number of groups after that.
#'
#' @details A grouping is a list of possible categories to group the data. The 'active group'
#' of these is the current grouping variables to be considered for analysis. The 'active group' can be any combination of the categories in a burst,
#' and can change with the use of `active_group()`.
#'
#' An `s_group` is a single row group. It is a 1xn dimensional list with any length(n) > 1.
#' Atleast one of the groups must be named 'id' which is the subjects id.
#'
#' A `c_grouping` is a collection of `s_groups`s, it is a data.frame with dimensions of 1xnrow(data).
#' One c_grouping has one 'active group' which describes the set of names in each s_group to group the data. When you change the 'active group', calculations and plots change accordingly
#' to the new grouping levels.
#'
#' You can create bursts with make_s_group and make_c_grouping.
#' @param x a list containing named grouping variables, one item must be named 'id'. ex: list(id = 1, month = 'may'). For a c_grouping: A list of s_groups or
#' a list of equal length named vectors which will be combined to create a c_grouping. ex: list(x = 1st_vector, y = 2nd_vector)
#' @param active_group a vector of the names of the groups to be considered 'active'.
#' @export make_c_grouping
#' @export make_s_group
#' @examples
#' # Make a single group
#' #'
#' make_s_group(x = list(id = "CJ11", month = 3, height = 10))
#'
#' # Make a c_grouping
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
#' mb1 <- make_c_grouping(x = burstz, active_group = c("id", "month"))
#' str(mb1)
#'
#' # Make a multi_burst from many ind_bursts
#' a <- make_s_group(list(id = 1, year = 2020))
#' b <- make_s_group(list(id = 1, year = 2020))
#' c <- make_s_group(list(id = 2, year = 2020))
#'
#' c(a, b, c)
#' @rdname grouping-class
make_s_group <- function(x) {
  # check duplicated
  if (any(duplicated(names(x)))) {
    stop("group names are duplicated")
  }

  # check id in burst
  check_group_id(x)

  new_group <- lapply(x, as.character)
  s_group(new_group)
}


s_group <- function(x = list(), active_group = NULL) {
  structure(
    x,
    class = c("s_group")
  )
}


c_grouping <- function(x = list(), active_group = NA) {
  if (is.na(active_group[1]) & length(x) > 0) {
    active_group <- names(x[[1]])
  }

  sort_index <- calc_sort_index(x, active_group)
  structure(x,
    active_group = active_group,
    sort_index = sort_index,
    class = c("c_grouping")
  )
}

#' @rdname grouping-class
make_c_grouping <-
  function(x = NULL,
           active_group = NULL) {
    if (inherits(x[[1]], c("s_group", "c_grouping"))) {
      group <- x
    } else {
      if (!inherits(x, "list")) {
        stop("grouping must be a list")
      }

      group <-
        do.call(function(...) {
          mapply(list, ..., SIMPLIFY = FALSE)
        }, x)
    }

    # to stop a problem where vector mode names are created via the first item in the list
    names(group) <- NULL
    if (is.null(active_group)) {
      active_group <- names(group[[1]])
    }
    check_NA_group(group)
    check_group_names(group)
    check_active_group(group, active_group, FALSE)
    check_two_groups(group, active_group = active_group)
    ret <- lapply(
      group,
      function(x, ...) {
        make_s_group(x)
      }
    )
    # ret
    cg <- c_grouping(ret, active_group = active_group)

    return(cg)
  }

#' @rdname grouping-class
#' @method c s_group
#' @export
#' @param ... objects to be pasted together into a c_grouping
c.s_group <- function(...) {
  ret <- list(...)
  # ret = list(ind_burst(list(id=1)), ind_burst(list(id=2)))
  unique_active_group(ret)
  make_c_grouping(ret)
}

#' @rdname grouping-class
#' @export
#' @method c c_grouping
#' @param recursive ignored
c.c_grouping <- function(..., recursive = FALSE) {
  x <- list(...)
  x <- unlist(x, recursive = FALSE)
  unique_active_group(x)
  active_group <- attr(x[[1]], "active_group")
  make_c_grouping(x, active_group = active_group)
}


#' @export
as.data.frame.c_grouping <- function(x, ...) {
  ret <- data.frame(row.names = seq_along(x))
  ret[[attr(x, "active_group")]] <- x
  ret
}

#' @export
print.s_group <- function(x, ...) {
  print(lapply(x, as.character))
}

#' @export
print.c_grouping <- function(x, ...) {
  print(lapply(x, function(x) {
    x
  }))
}

#' @export
#' @importFrom utils str
str.c_grouping <- function(object, ...) {
  n <- length(object)
  cat(paste0(class(object)[1], " object \n"))
  utils::str(unclass(object), list.len = 1)
}


#' @export
format.s_group <- function(x, ...) {
  message(paste0("(", paste0(
    names(x), ": ", as.character(unlist(x)),
    collapse = ", "
  ), ")"))
}

#' @export
format.c_grouping <- function(x, ...) {
  paste0("(", vapply(x, function(y) {
    paste(
      paste0(names(y), ": ", as.character(unlist(y))),
      collapse = ", "
    )
  }, NA_character_), ")")
}

#' @export
"[.c_grouping" <- function(x, i, j, ...) {
  nargs <- nargs()
  if (!missing(i) && is.character(i)) {
    i <- which(attr(x, "sort_index") %in% i)
  }
  active_group <- attr(x, "active_group")
  c_grouping(NextMethod(), active_group = active_group)
}

#' @export
"[<-.c_grouping" <- function(x, i, value) {
  if (is.null(value) || inherits(value, "s_group")) {
    value <- list(value)
  }
  # attributes(x[[1]])
  unique_active_group(list(x[[1]], value[[1]]))
  x <- unclass(x) # becomes a list, but keeps attributes
  ret <- c_grouping(NextMethod())
  check_group_names(ret)
  check_two_groups(ret)
  check_NA_group(ret)
  ret
}

#' @export
"[<-.s_group" <- function(x, i, value) {
  if (is.null(value) || inherits(value, "s_group")) {
    value <- list(value)
  }
  x <- unclass(x) # becomes a list, but keeps attributes

  ret <- make_s_group(NextMethod())
  structure(ret)
}

#' @export
"$<-.s_group" <- function(x, i, value) {
  if (is.null(value) || inherits(value, "s_group")) {
    value <- list(value)
  }
  x <- unclass(x) # becomes a list, but keeps attributes
  ret <- make_s_group(NextMethod())
  structure(ret)
}

#' @export
summary.c_grouping <- function(object, ...) {
  ab <- paste0(attr(object, "active_group"), collapse = ", ")
  object <- group_labels(object)
  levels(object) <- c(levels(object), paste("active_group:", ab))
  NextMethod()
}

#' @title Shows grouping labels created from the s_group and the c_grouping
#' @name group_labels
#' @param x a sftrack or grouping object
#' @export
group_labels <- function(x) {
  UseMethod("group_labels", object = x)
}


#' @export
#' @rdname group_labels
group_labels.sftrack <- function(x) {
  attr(x[[attr(x, "group_col")]], "sort_index")
}

#' @export
#' @rdname group_labels
group_labels.sftraj <- function(x) {
  attr(x[[attr(x, "group_col")]], "sort_index")
}


#' @export
#' @rdname group_labels
group_labels.c_grouping <- function(x) {
  attr(x, "sort_index")
}

#' @title Access the active_group value
#' @rdname active_group
#' @param x a c_grouping
#' @export
#' @description The active group is the combination of group names to group the data sets.
#' The active_group acts essentially like a paste(names_of_groups, sep = '_') grouping variable.
#' @examples
#' #'
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
#' mb1 <- make_c_grouping(x = burstz, active_group = c("id", "month"))
#'
#' # see the current active burst
#' active_group(mb1)
#'
#' # change the active burst
#' active_group(mb1) <- "id"
#'
#' # Using a full data set
#' my_track <- as_sftrack(raccoon,
#'   time = "timestamp",
#'   error = NA, coords = c("longitude", "latitude"),
#'   group = burstz
#' )
#'
#' summary(my_track)
#'
#' # change active group
#' active_group(my_track$sft_group) <- "id"
#'
#' summary(my_track)
active_group <- function(x) {
  if (inherits(x, c("sftrack", "sftraj"))) {
    return(attr(x[[attr(x, "group_col")]], "active_group"))
  }
  # check_burst_names(x)
  attr(x, "active_group")
}

#' @title Set new active group
#' @export
#' @rdname active_group_replace
#' @param x sftrack/sftraj/c_grouping/s_group object
#' @param value character vector of the grouping names to make active
"active_group<-" <- function(x, value) {
  UseMethod("active_group<-", object = x)
}

#' @export
#' @name active_group_replace
"active_group<-.sftrack" <- function(x, value) {
  # x = my_sftrack
  # value = 'id'
  group <- x[[attr(x, "group_col")]]

  if (!all(value %in% names(group[[1]]))) {
    stop("not all new active group found in current grouping names")
  }

  attr(x[[attr(x, "group_col")]], "active_group") <- value
  attr(x[[attr(x, "group_col")]], "sort_index") <- calc_sort_index(x[[attr(x, "group_col")]])
  structure(x)
}

#' @export
#' @name active_group_replace
"active_group<-.sftraj" <- function(x, value) {
  group <- x[[attr(x, "group_col")]]

  if (!all(value %in% names(group[[1]]))) {
    stop("not all new active group found in current grouping names")
  }

  attr(x[[attr(x, "group_col")]], "active_group") <- value
  attr(x[[attr(x, "group_col")]], "sort_index") <- calc_sort_index(x[[attr(x, "group_col")]])
  step_recalc(x)
}

#' @export
#' @name active_group_replace
"active_group<-.c_grouping" <- function(x, value) {
  # x = my_sftrack
  # value = 'id'
  group <- x

  if (!all(value %in% names(group[[1]]))) {
    stop("not all new group names found in grouping burst")
  }

  attr(x, "active_group") <- value
  attr(x, "sort_index") <- calc_sort_index(x)
  structure(x)
}

#' @title Calculate a new sort index for groups
#' @param x group or sftrack object
#' @param active_group (optional), a new active group. If not included, defaults  to the active group (if a c_grouping) or the group names
#' @export
calc_sort_index <- function(x, active_group = NA) {
  if (inherits(x, "c_grouping")) {
    active_group <- attr(x, "active_group")
  }
  if (any(is.na(active_group))) {
    active_group <- names(x[[1]])
  }

  factor(vapply(x, function(y) {
    paste0(y[active_group], collapse = "_")
  }, character(1)))
}

#' @title Display the levels of the sort index
#' @export
#' @rdname group_name
#' @param x sftrack/sftraj/c_grouping/s_group object
group_names <- function(x) {
  UseMethod("group_names", object = x)
}

#' @export
#' @name group_name
group_names.c_grouping <- function(x) {
  levels(attr(x, "sort_index"))
}

#' @export
#' @name group_name
group_names.sftraj <- function(x) {
  levels(attr(x[[attr(x, "group_col")]], "sort_index"))
}

#' @export
#' @name group_name
group_names.sftrack <- function(x) {
  levels(attr(x[[attr(x, "group_col")]], "sort_index"))
}

test_that("s_group works", {
  ind <- make_s_group(list(id = 1, month = "Jan"))
  expect_equal(names(ind), c("id", "month"))

  # burst names not duplicated
  expect_error(make_s_group(list(id = 1, id = 1)), "group names are duplicated")

  # There is an id column
  expect_error(make_s_group(list(month = "Jan", Year = 2020)), "There is no `id` column in group names")
})

test_that("c_grouping works", {
  burstz <- list(id = rep(1, 4), col1 = c(1, 1, 2, 2), col2 = c("A", "B", "B", "A"))

  obj1 <- suppressWarnings(make_c_grouping(x = burstz, active_group = "id"))
  expect_equal(sapply(obj1, function(x) x$col1), as.character(c("1", "1", "2", "2")))

  cg <- make_c_grouping(x = burstz, active_group = "id")
  expect_equal(attr(cg, "active_group"), "id")
  expect_message(make_c_grouping(x = burstz))

  # check levels
})

test_that("concatenate groups works", {
  mb1 <- make_s_group(list(id = 1, month = "Jan"))
  mb2 <- make_s_group(list(id = 1, month = "Feb"))
  comb <- c(mb1, mb1, mb2, mb2)
  expect_equal(attr(comb, "active_group"), c("id", "month"))

  burst1 <- make_c_grouping(list(id = rep(1, 4), col1 = c(1, 1, 2, 2)))
  burst2 <- make_c_grouping(list(id = rep(1, 4), col1 = c("A", "B", "B", "A")))

  comb <- c(burst1, burst2)
  expect_equal(attr(comb, "active_group"), c("id", "col1"))

  burst1 <- make_c_grouping(list(id = rep(1, 4), col1 = c(1, 1, 2, 2)))
  burst2 <- make_c_grouping(list(id = rep(1, 4), col2 = c("A", "B", "B", "A")))
  expect_error(c(burst1, burst2), "Group names do not match")
})

test_that("active group can change correctly", {
  burst1 <- make_c_grouping(list(id = rep(1, 4), col2 = c("A", "B", "B", "A")))
  expect_equal(active_group(burst1), c("id", "col2"))
  active_group(burst1) <- "id"
  expect_equal(active_group(burst1), c("id"))

  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = c(0, 1, 2, 3),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:4)
  )
  my_sftrack <- as_sftrack(
    data = df1, group = c("id", "month"),
    time = "timez", active_group = c("id", "month"), coords = c("x", "y")
  )
  expect_equal(active_group(my_sftrack), c("id", "month"))
  active_group(my_sftrack) <- "id"
  expect_equal(active_group(my_sftrack), c("id"))
})

test_that("subset c_grouping", {
  burst1 <- make_c_grouping(list(id = rep(1, 4), col2 = c("A", "B", "B", "A")))
  expect_equal(class(burst1[1]), "c_grouping")

  # subsetting via label name
  expect_equal(length(burst1["1_B"]), 2)

  # replace item in a multi_burst
  burst1 <- make_c_grouping(list(id = rep(1, 5), col2 = c("A", "A", "C", "C", "A")))
  burst1[2] <- s_group(list(id = 1, col2 = "C"))
  expect_equal(class(burst1[[2]]), "s_group")
  expect_equal(as.character(attr(burst1, "sort_index")[2]), "1_C")

  # replacing an individual element in a multi_burst
  burst1[[1]]$id <- 3
  expect_equal(burst1 [[1]][[1]], "3")


  # subset to 0
  mb_attr <- attributes(burst1[F])
  expect_equal(mb_attr, list(active_group = c("id", "col2"), sort_index = factor(NULL), class = "c_grouping"))
})

test_that("sftraj is built correct", {
  # data.frame mode works successfully
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = c(0, 1, 2, 3),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:4)
  )
  my_sftraj <- as_sftraj(
    data = df1, group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id", "month"), coords = df1[, c("x", "y")]
  )
  expect_equal(colnames(my_sftraj), c("id", "month", "x", "y", "z", "timez", "sft_timestamp", "sft_group", "geometry"))
  expect_equal(class(my_sftraj$geometry)[1], "sfc_GEOMETRY")

  # vector mode with data

  my_sftraj <- as_sftraj(
    data = df1, group = c("id", "month"),
    time = "timez", active_groupo = c("id", "month"), coords = c("x", "y")
  )
  expect_equal(colnames(my_sftraj), c("id", "month", "x", "y", "z", "timez", "sft_group", "geometry"))

  # data.frame mode without data; accepts null
  my_sftraj <- as_sftraj(
    group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id", "month"), coords = df1[, c("x", "y")]
  )
  expect_equal(colnames(my_sftraj), c("sftrack_id", "sft_timestamp", "sft_group", "geometry"))

  # check 2 dimensions
  expect_equal(class(my_sftraj$geometry[[1]])[1], "XY")

  # test that sftraj can change active_bursts
  expect_equal(attr(my_sftraj$sft_group, "active_group"), c("id", "month"))
  my_sftraj <- suppressMessages(as_sftraj(
    group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id"), coords = df1[, c("x", "y", "z")]
  ))
  expect_equal(attr(my_sftraj$sft_group, "active_group"), c("id"))

  # test that dimensions are created equally
  expect_equal(class(my_sftraj$geometry[[1]])[1], "XYZ")


  # Correctly makes lines and points
  df1 <- data.frame(
    id = c(1, 1, 1, 1, 1, 1),
    month = c(1, 1, 1, 1, 1, 1),
    x = c(27, 27, 27, NA, 29, 30),
    y = c(-80, -81, -82, NA, 83, 83),
    z = c(0, 1, 2, 3, 4, 5),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:6)
  )
  my_sftraj <- as_sftraj(
    data = df1, group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id", "month"), coords = df1[, c("x", "y")]
  )
  expect_equal(
    unlist(my_sftraj$geometry),
    c(27, 27, -80, -81, 27, 27, -81, -82, 27, -82, NA, NA, 29, 30, 83, 83, 30, 83)
  )

  zerotraj <- my_sftraj[F, ]
  expect_equal(nrow(zerotraj), 0)
})

test_that("as_sftraj and sftraj convert back and forth successfully", {
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = c(0, 1, 2, 3),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:4)
  )
  my_sftrack <- as_sftrack(
    data = df1, group = c("id", "month"),
    time = "timez", active_group = c("id", "month"), coords = c("x", "y")
  )
  my_sftraj <- as_sftraj(
    data = df1, group = c("id", "month"),
    time = "timez", active_group = c("id", "month"), coords = c("x", "y")
  )
  new_sftraj <- as_sftraj(my_sftrack)
  expect_equal(new_sftraj, my_sftraj)
  new_sftrack <- as_sftrack(my_sftraj)
  expect_equal(new_sftrack, my_sftrack)
})

test_that("input as sf successfully", {
  # Input is a sf object

  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = c(0, 1, 2, 3),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:4)
  )

  sf_df <- st_as_sf(df1, coords = c("x", "y"))
  new_sftraj <- as_sftraj(data = sf_df, group = "id", time = "timez")

  # Not include sfc_point
  sf_df <- st_as_sf(df1, coords = c("x", "y"))
  sf_df$geometry <- st_sfc(list(st_multipoint(), st_multipoint(), st_multipoint(), st_multipoint()))

  expect_error(as_sftraj(data = sf_df, group = "id", time = "timez"))
})


test_that("subset works correctly", {
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = c(0, 1, 2, 3),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60 * 60 * (1:4)
  )
  # retains sftraj class
  my_sftraj <- as_sftraj(
    data = df1, group = c("id", "month"),
    time = "timez", active_group = c("id", "month"), coords = c("x", "y")
  )
  expect_equal(class(my_sftraj[1:3, ]), c("sftraj", "sf", "data.frame"))

  expect_equal(ncol(my_sftraj[, 3, ]), 4)


  expect_silent(my_sftraj[, 3, drop = T])

  expect_equal(nrow(my_sftraj[, 3, drop = T]), NULL)

  # subset by colname without dropped columns
  expect_equal(colnames(my_sftraj[, c("id", "month")]), c("id", "month", "sft_group", "timez", "geometry"))


  # rbind
  df2 <- df1
  df2$timez <- df2$timez + 10
  my_sftraj2 <- as_sftraj(
    data = df2, group = c("id", "month"),
    time = "timez", active_group = c("id", "month"), coords = c("x", "y")
  )
  my_sftraj3 <- rbind(my_sftraj, my_sftraj2)
  expect_equal(class(my_sftraj3)[1], "sftraj")

  expect_equal(nrow(my_sftraj3), 8)

  # change active burst
  active_group(my_sftraj2) <- "id"
  expect_equal(attr(my_sftraj2$sft_group, "active_group"), "id")

  # check that geometries were recalculated
  active_group(my_sftraj2)
  unlist(my_sftraj2$geometry)
  expect_equal(unlist(my_sftraj2$geometry), c(27, 27, -80, -81, 27, 27, -81, -82, 27, 27, -82, -83, 27, -83))
})

#' Convert objects into sftrack objects.
#' @name as_sftrack
#' @title Convert objects into sftrack objects.
#' @description
#' This function converts x,y,z data into an sftrack object with a sf_geometry column of sf_POINTS.
#' Creates a `grouping` column to group movement data and sets dedicated time and error columns.
#'
#' Raw data inputted in two ways: vector or data.frame. 'Vector' inputs gives the argument as a vector where
#' length = nrow(data). 'Data.frame' inputs gives the arguments as the column name of `data` where the input can be found.
#' Either input is allowed on any given argument.
#'
#' Some options are global and required regardless
#' @param data a data.frame of the movement data, if supplied all data.frame inputs, than is optional
#' @param coords a character vector describing where the x,y,z coordinates are located in `data` or a list with x,y,z (optional) vectors
#' @param group a list of named vectors describing multiple grouping variables or  a character vector naming the other grouping columns in `data`.
#' @param active_group a character vector of the burst names to be 'active' to group data by for analysis
#' @param time a vector of time information, can be either POSIX or an integer or a character string naming the column in `data` where the time information is located
#' @param error (optional) a vector of error information for the movement dataa character string naming the column in `data` where the error information is located
#' @param crs Coordinate reference system to be assigned; object of class `crs`. Defaults to NA
#' @param zeroNA logical whether to convert 0s in spatial data into NAs. Defaults to FALSE.
#' @param group_name (optional) new column name for grouping data
#' @param timestamp_name (optional) new column name for time data
#' @param error_name (optional) new column name for error data
#' @param overwrite_names T/F Whether to overwrite data if a group/time/error column name is supplied but already in data
#' @param ... extra information to be passed on to as_sftrack
#' @import sf
#' @export
#' @examples
#' #'
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
#' # Input is a data.frame
#' my_track <- as_sftrack(raccoon,
#'   group = burstz, time = "timestamp",
#'   error = NA, coords = c("longitude", "latitude")
#' )
#'
#' # Input is a ltraj
#' library("adehabitatLT")
#' ltraj_df <- as.ltraj(
#'   xy = raccoon[, c("longitude", "latitude")],
#'   date = as.POSIXct(raccoon$timestamp),
#'   id = raccoon$animal_id, typeII = TRUE,
#'   infolocs = raccoon[, 1:6]
#' )
#'
#' my_sftrack <- as_sftrack(ltraj_df)
#' head(my_sftrack)
#'
#' # Input is a sf object
#' library("sf")
#' df1 <- raccoon[!is.na(raccoon$latitude), ]
#' sf_df <- st_as_sf(df1, coords = c("longitude", "latitude"))
#'
#' new_sftrack <- as_sftrack(sf_df, group = c(id = "animal_id"), time = "timestamp")
#' head(new_sftrack)
#'
#' # Input is an sftraj object
#' my_traj <- as_sftraj(raccoon,
#'   time = "timestamp",
#'   error = NA, coords = c("longitude", "latitude"),
#'   group = burstz
#' )
#'
#' new_track <- as_sftrack(my_traj)
#' head(new_track)
#' ######################
#' # Builder
as_sftrack <- function(data = data.frame(), ...) {
  UseMethod("as_sftrack", object = data)
}

#' @title Define an sftrack
#' @param data  data.frame with multi_burst column, geometry column, time_col (integer/POSIXct), and error column (optional)
#' @param group_col column name of grouping info in `data`
#' @param sf_col column name of geometry info in `data`
#' @param time_col column name of time info in `data`
#' @param error_col column name of error info in `data`
#' @export
new_sftrack <-
  function(data,
           group_col,
           sf_col,
           time_col,
           error_col = NA) {
    data_sf <- sf::st_sf(data, sf_column_name = sf_col)
    structure(
      data_sf,
      group_col = group_col,
      time_col = time_col,
      error_col = error_col,
      class = c("sftrack", "sf", "data.frame")
    )
  }

#' @rdname as_sftrack
#' @export
#' @method as_sftrack data.frame
as_sftrack.data.frame <- function(data = data.frame(),
                                  ...,
                                  coords = c("x", "y"),
                                  group = "id",
                                  active_group = NA,
                                  time = "time",
                                  error = NA,
                                  crs = NA,
                                  zeroNA = FALSE,
                                  group_name = "sft_group",
                                  timestamp_name = "sft_timestamp",
                                  error_name = "sft_error",
                                  overwrite_names = FALSE) {
  # data = data.frame()
  # coords = sub_gps[,c('longitude','latitude')]
  # burst = list(id=sub_gps$id, numSat = sub_gps$numSat)
  # time = sub_gps[,'timez']
  # crs=4326
  # error = NA
  # zeroNA = FALSE
  # active_burst = NA
  ######################
  # Check inputs
  # Id
  if (nrow(data) == 0) {
    data <- data.frame(sftrack_id = seq_along(time))
  }
  # bursts
  if (length(group) == 1) {
    names(group) <- "id"
  }
  if (all(sapply(group, length) == nrow(data))) {
    # check id in burst
    check_group_id(group)
    group_list <- group
  } else {
    # check names exist
    if (inherits(group, "list")) {
      group <- vapply(group, c, character(1))
    }
    check_names_exist(data, group)
    # check id in burst
    # check id in burst
    check_group_id(group)
    # create burst list from names
    group_list <- lapply(data[, group, FALSE], function(x) {
      x
    })
    if (!is.null(names(group))) {
      names(group_list) <-
        names(group)
    } else {
      names(group_list) <- group
    }
  }

  # Coords
  if (is.null(nrow(coords))) {
    check_names_exist(data, coords)
    xyz <- data[, coords]
  } else {
    xyz <- as.data.frame(coords)
  }
  # fix zeros to NA
  if (zeroNA) {
    xyz <- fix_zero(xyz)
  }
  check_NA_coords(xyz)

  # Time
  if (length(time) == nrow(data)) {
    if (timestamp_name %in% colnames(data) && !overwrite_names) {
      stop(paste0("column name: \"", timestamp_name, "\" already found in data.frame.
If youd like to overwrite column use overwrite_names = TRUE"))
    } else {
      data[[timestamp_name]] <- time
      time_col <- timestamp_name
    }
  } else {
    check_names_exist(data, time)
    time_col <- time
  }
  check_time(data[[time_col]])

  # Error
  if (length(error) == nrow(data)) {
    # Decide whether to overwrite names or not
    if (error_name %in% colnames(data) && !overwrite_names) {
      stop(paste0("column name: \"", error_name, "\" already found in data.frame.
If youd like to overwrite column use overwrite_names = TRUE"))
    } else {
      data[[error_name]] <- error
    }
    error_col <- error_name
  } else {
    if (!is.na(error)) {
      check_names_exist(data, error)
      error_col <- error
    } else {
      error_col <- NA
    }
  }

  # pull out other relevant info
  if (any(is.na(active_group))) {
    active_group <- names(group_list)
  }
  group <-
    make_c_grouping(group_list, active_group = active_group)

  # earliest reasonable time to check time stamps
  dup_timestamp(time = data[[time_col]], x = group)

  geom <-
    sf::st_as_sf(xyz,
      coords = names(xyz),
      crs = crs,
      na.fail = FALSE
    )
  # Force calculation of empty geometries.
  attr(geom[, attr(geom, "sf_column")], "n_empty") <-
    sum(vapply(st_geometry(geom), sfg_is_empty, TRUE))

  # Decide whether to overwrite names or not
  if (group_name %in% colnames(data) && !overwrite_names) {
    stop(paste0("column name: \"", group_name, "\" already found in data.frame.
If youd like to overwrite column use overwrite_names = TRUE"))
  } else {
    data[[group_name]] <- group
  }
  data$geometry <- st_geometry(geom)

  ret <- new_sftrack(
    data = data,
    group_col = group_name,
    sf_col = "geometry",
    error_col = error_col,
    time_col = time_col
  )
  # Sanity checks
  ret <- ret[check_ordered(ret[[attr(ret, "group_col")]], ret[[attr(ret, "time_col")]]), ]

  check_z_coords(ret)

  return(ret)
}

#' @rdname as_sftrack
#' @method as_sftrack sftraj
#' @export
as_sftrack.sftraj <- function(data, ...) {
  geometry <- st_geometry(data)

  # pull out first points from straj
  new_geom <- pts_traj(data)
  crs <- attr(geometry, "crs")

  geometry <- sf::st_sfc(new_geom, crs = crs)
  group <- attr(data, "group_col")
  error <- attr(data, "error")
  time <- attr(data, "time")
  sf_col <- attr(data, "sf_column")
  data[[sf_col]] <- geometry

  new_data <- as.data.frame(data)
  ret <- new_sftrack(
    data = new_data,
    group_col = group,
    sf_col = sf_col,
    error_col = error,
    time_col = time
  )

  return(ret)
}

### Ltraj
#' @rdname as_sftrack
#' @method as_sftrack ltraj
#' @export
as_sftrack.ltraj <- function(data, ...) {
  # This is done so we dont have to import adehabitat. (instead of ld())
  # But it could go either way depending
  new_data <- lapply(seq_along(data), function(x) {
    sub <- data[x, ]
    attributes(sub[[1]])
    id <- attr(sub[[1]], "id")
    burst <- attr(sub[[1]], "burst")
    infolocs <- infolocs(data)[x]
    sft_timestamp <- sub[[1]]$date
    coords <- c("x", "y")
    data.frame(sub[[1]][, coords], id, burst, sft_timestamp, infolocs)
  })
  df1 <- do.call(rbind, new_data)
  time <- "sft_timestamp"
  group <- list(id = df1$id)
  crs <- attr(data, "proj4string")
  # pull out id and burst from ltraj object
  id_lt <- vapply(data, function(x) {
    attr(x, "id")
  }, NA_character_)
  group_lt <-
    vapply(data, function(x) {
      attr(x, "burst")
    }, NA_character_)

  if (!all(group_lt == id_lt)) {
    group$group <- df1$burst
  }
  coords <- c("x", "y")
  geom <-
    sf::st_as_sf(df1[, coords],
      coords = coords,
      crs = crs,
      na.fail = FALSE
    )
  # pull out other relevant info
  df1$sft_group <- make_c_grouping(group)
  error <- NA
  new_data <-
    cbind(df1[, !colnames(df1) %in% c("id")], geometry = st_geometry(geom))
  ret <- new_sftrack(
    data = new_data,
    group_col = "sft_group",
    error_col = error,
    time_col = time,
    sf_col = "geometry"
  )
  # Sanity check. Which are necessary?
  ret <-
    ret[check_ordered(ret[[attr(ret, "group_col")]], ret[[attr(ret, "time_col")]]), ]
  #
  return(ret)
}
# sf
#' @rdname as_sftrack
#' @method as_sftrack sf
#' @export
as_sftrack.sf <- function(data,
                          ...,
                          coords,
                          group,
                          active_group = NA,
                          time,
                          error = NA,
                          group_name = "sft_group",
                          timestamp_name = "sft_timestamp",
                          error_name = "sft_error",
                          overwrite_names = FALSE) {
  geom <- st_geometry(data)

  data <- as.data.frame(data)
  # Check inputs
  # geom
  if (attributes(geom)$class[1] != "sfc_POINT") {
    stop("sf column must be an sfc_POINT")
  }
  # Id
  if (nrow(data) == 0) {
    data <- data.frame(sftrack_id = seq_along(time))
  }
  # bursts
  if (length(group) == 1) {
    names(group) <- "id"
  }
  if (all(sapply(group, length) == nrow(data))) {
    # check id in burst
    check_group_id(group)
    group_list <- group
  } else {
    # check names exist
    if (inherits(group, "list")) {
      group <- vapply(group, c, character(1))
    }
    check_names_exist(data, group)
    # check id in burst
    # check id in burst
    check_group_id(group)
    # create burst list from names
    group_list <- lapply(data[, group, FALSE], function(x) {
      x
    })
    if (!is.null(names(group))) {
      names(group_list) <-
        names(group)
    } else {
      names(group_list) <- group
    }
  }

  # Time
  if (length(time) == nrow(data)) {
    if (timestamp_name %in% colnames(data) && !overwrite_names) {
      stop(paste0("column name: \"", timestamp_name, "\" already found in data.frame.
If youd like to overwrite column use overwrite_names = TRUE"))
    } else {
      data[[timestamp_name]] <- time
      time_col <- timestamp_name
    }
  } else {
    check_names_exist(data, time)
    time_col <- time
  }
  check_time(data[[time_col]])

  # Error
  if (length(error) == nrow(data)) {
    # Decide whether to overwrite names or not
    if (error_name %in% colnames(data) && !overwrite_names) {
      stop(paste0("column name: \"", error_name, "\" already found in data.frame.
If youd like to overwrite column use overwrite_names = TRUE"))
    } else {
      data[[error_name]] <- error
    }
    error_col <- error_name
  } else {
    if (!is.na(error)) {
      check_names_exist(data, error)
      error_col <- error
    } else {
      error_col <- NA
    }
  }
  #
  if (any(is.na(active_group))) {
    active_group <- names(group_list)
  }
  group <-
    make_c_grouping(group_list, active_group = active_group)

  # earliest reasonable time to check time stamps
  dup_timestamp(time = data[[time_col]], x = group)

  # Decide whether to overwrite names or not
  if (group_name %in% colnames(data) && !overwrite_names) {
    stop(paste0("column name: \"", group_name, "\" already found in data.frame.
If youd like to overwrite column use overwrite_names = TRUE"))
  } else {
    data[[group_name]] <- group
  }

  ret <- new_sftrack(
    data = data,
    group_col = group_name,
    sf_col = "geometry",
    error_col = error_col,
    time_col = time_col
  )
  # Sanity checks
  ret <- ret[check_ordered(ret[[attr(ret, "group_col")]], ret[[attr(ret, "time_col")]]), ]

  check_z_coords(ret)

  return(ret)
}


# Methods for 'sftrack' class

#' @title Print methods for sftrack
#' @name Print_sftrack_objects
#' @param x sftraj object
#' @param n_row Integer of number of rows to display. Defaults to global option default if non supplied
#' @param n_col Integer of number of columns to display + required sftrack columns (burst, geometry, time, and error). Defaults to global option default if non supplied
#' @param ... other arguments passed onto print
#' @export
print.sftrack <- function(x, n_row, n_col, ...) {
  if (missing(n_col)) {
    n_col <- ncol(x)
  }
  if (missing(n_row)) {
    n_row <- nrow(x)
  }
  group_col <- attr(x, "group_col")
  sf_col <- attr(x, "sf_column")
  time_col <- attr(x, "time_col")
  sf_attr <- attributes(sf::st_geometry(x))
  # time stuff

  tcl <- class(x[[time_col]])[1]
  if (tcl == "POSIXct") {
    tz <- attributes(x[[time_col]])$tzone
    if (is.null(tz) || tz == "") {
      tz <- paste("no timezone")
    }
    time_mes <- paste(tcl, "in", tz)
  } else {
    time_mes <- "integer"
  }

  # active burst
  ab <- attr(x[[group_col]], "active_group")
  bn <- names(x[[group_col]][[1]])
  active_group_names <- paste0("*", ab, "*")
  group_mes <-
    paste0(active_group_names, bn[!ab %in% bn], collapse = ", ")
  # print
  cat(
    paste0(
      "Sftrack with ",
      nrow(x),
      " features and ",
      ncol(x),
      " fields (",
      sf_attr$n_empty,
      " empty geometries) \n"
    )
  )
  cat(paste0(
    'Geometry : \"',
    sf_col,
    '\" (',
    class(x[[sf_col]][[1]])[1],
    ", crs: ",
    format(sf_attr$crs),
    ") \n"
  ))
  cat(paste0('Timestamp : \"', time_col, '\" (', time_mes, ") \n"))
  cat(paste0('Groupings : \"', group_col, '\" (', group_mes, ") \n"))
  cat("-------------------------------\n")
  # Figure out the row and columns
  row_l <- ifelse(nrow(x) > n_row, n_row, nrow(x))
  sub_col_names <-
    colnames(x)[!colnames(x) %in% c(group_col, sf_col, time_col)]
  col_l <- length(sub_col_names)

  x <- as.data.frame(x)

  if (n_col < col_l) {
    p <- ifelse(col_l > n_col, n_col, col_l)
    ret <- cbind(
      x[1:row_l, sub_col_names[1:p], drop = FALSE],
      data.frame("..." = rep("...", row_l)),
      x[1:row_l, c(group_col, sf_col, time_col)]
    )
  } else {
    ret <- x
  }

  ret <- ret[1:row_l, ]
  print(ret, ...)
}
# print(my_track,10,10)

# Sumary
# summary.sftrack
#' @export
summary.sftrack <- function(object, ..., stats = FALSE) {
  if (stats) {
    summary_sftrack(object)
  } else {
    (NextMethod())
  }
}
# summary(my_sftrack,stats=TRUE)
#' @export
`[.sftrack` <- function(x, i, j, ..., drop = FALSE) {
  # x = my_sftrack
  # i = 1:10
  # j=c(1,2,3)
  # rm(j)
  group_col <- attr(x, "group_col")
  sf_col <- attr(x, "sf_column")
  time_col <- attr(x, "time_col")
  error_col <- attr(x, "error_col")
  # if(is.na(error_col)){ error_col <- NULL}
  nargs <- nargs()
  # print(nargs)
  if (!missing(j) && missing(i)) {
    i <- seq_len(nrow(x))
  }
  if (!missing(i) && nargs > 2) {
    if (is.character(i)) {
      i <- group_labels(x) %in% i
    }
  }
  if (!missing(j) && all(is.character(j))) {
    j <- which(colnames(x) %in% j)
  }
  # x = as.data.frame(x)
  class(x) <- setdiff(class(x), c("sftrack", "sf"))
  x <- if (missing(j)) {
    if (nargs == 2) {
      x[i]
    } else {
      x[i, , ]
    }
  } else {
    if (drop) {
      return(x[i, j, drop = drop])
    } else {
      error_col <- if (is.na(error_col)) {
        NULL
      } else {
        error_col
      }
      x[i, union(colnames(x)[j], c(group_col, sf_col, time_col, error_col))]
    }
  }


  # # This piece is if we want drop = TRUE to be the default
  # if((any(!exist_name[1:3]) || !is.na(error_col) & !exist_name[4])){
  #
  #   message(paste0(paste0(c('burst','geometry','time','error')[!exist_name],collapse=', '),' subsetted out of sftrack object, reverting to ',class(x)[1],
  #     '\n Use drop = FALSE to retain class'))
  #
  #     }
  ret <- new_sftrack(
    x,
    group_col = group_col,
    sf_col = sf_col,
    time_col = time_col,
    error_col = error_col
  )
  dup_timestamp(ret)
  ret
}

#' @export
rbind.sftrack <- function(...) {
  all <- list(...)
  # all = list(my_sftrack, my_sftrack2)

  same_att <-
    all(duplicated(lapply(all, function(x) {
      attributes(x)[c("sf_column", "time", "error")]
    }))[-1])
  if (!same_att) {
    stop("sf, time, and error columns must be the same")
  }
  att <- attributes(all[[1]])
  time_col <- att$time_col
  error_col <- att$error_col
  group_col <- att$group_col
  sf_col <- att$sf_column
  for (i in seq_along(all)) {
    class(all[[i]]) <- setdiff(class(all[[i]]), c("sftrack", "sf"))
  }
  new_df <- do.call(rbind, all)
  class(new_df) <- setdiff(class(new_df), c("sftrack", "sf"))
  ret <- new_sftrack(
    data = new_df,
    group_col = group_col,
    time_col = time_col,
    error_col = error_col,
    sf_col = sf_col
  )

  # Sanity checks
  dup_timestamp(ret)
  return(ret)
}

#' @export
cbind.sftrack <- function(..., deparse.level = 1) {
  all <- list(...)
  # all <- list(my_sftrack,my_sftrack[,1:5,drop=T])
  sftrack_obj <- sapply(all, function(x) {
    inherits(x, "sftrack")
  })
  if (sum(sftrack_obj) > 1) {
    # return error
    stop("Cannot attempt to merge two sftrack objects together, use new_sftrack() instead")
  }
  # The rest below this is not useful as I'm pretty sure you cbind.sftrack only works when both objects are an sftrack object
  # I'll keep it till we can decide what we want cbind to do
  # check if theres two burst columns
  if (sum(unlist(lapply(all, function(x) {
    colnames(x) == attr(x, "group_col")
  }))) > 1) {
    stop("More than one group column found")
  }
}

#' @export
"[[<-.sftrack" <- function(x, i, value) {
  x <- structure(NextMethod(), class = c("sftrack", "sf", "data.frame"))
}

test_that("checks check successfully", {

  # columns not found in data set
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = 0:3,
    timez = Sys.time() + 60 * 60 * (c(1, 2, 3, 4))
  )
  expect_error(as_sftrack(
    data = df1, group = list(id = df1$id, month = df1$month),
    time_col = "nottime", active_group = c("id", "month"), coords = c("x", "y", "z")
  ))

  # Z coordinates
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = 0:3,
    timez = Sys.time() + 60 * 60 * (1:4)
  )
  expect_message(
    as_sftrack(
      data = df1, group = list(id = df1$id, month = df1$month),
      time = "timez", active_group = c("id", "month"), coords = c("x", "y", "z")
    ),
    "Z coordinates found. Note that the vast majority of rgdal calculations are done using 2D geometry"
  )

  # duplicate time_stamps
  df2 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c("march", "march", "april", "april"),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    timez = Sys.time() + 60 * 60 * c(1, 1, 2, 3)
  )
  expect_error(
    as_sftrack(
      data = df2, group = c("id", "month"),
      time = "timez", coords = c("x", "y")
    ),
    "groups: 1_march have duplicated time stamps"
  )

  # NAs in some coordinate columns
  df3 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = c(NA, NA, 1, 0),
    timez = Sys.time() + 60 * 60 * (1:4)
  )
  expect_error(
    as_sftrack(
      data = df3, group = c("id", "month"),
      time = "timez", coords = c("x", "y", "z")
    ),
    "z column has NAs that are not found in other coordinate columns"
  )

  # check coord names exist
  expect_error(
    as_sftrack(
      data = df1, group = c("id", "month"),
      time = "timez", coords = c("x", "y", "z1")
    ),
    "z1 not found in data frame"
  )

  # data not ordered
  df4 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    timez = Sys.time() + 60 * 60 * c(4, 1, 2, 3)
  )
  expect_message(
    as_sftrack(
      data = df4, group = c("id", "month"),
      time = "timez", coords = c("x", "y")
    ),
    "time was not ordered for each burst"
  )
})

test_that("sftraj conversions", {
  # multiple bursts become combined
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = 0:3,
    timez = Sys.time() + 60 * 60 * (1:4)
  )
  my_sftraj <- suppressMessages(as_sftraj(
    data = df1, group = c("id", "month"),
    time = "timez", coords = c("x", "y", "z")
  ))
  #  my_sftraj
  # sftraj -> sftrack -> sftraj
  new_sftrack <- as_sftrack(my_sftraj)
  # new_sftrack
  conv_back_sftraj <- as_sftraj(new_sftrack)
  expect_equal(my_sftraj, conv_back_sftraj)
})


test_that("sftrack conversions", {
  # multiple bursts become combined
  df1 <- data.frame(
    id = c(1, 1, 1, 1),
    month = c(1, 1, 2, 2),
    x = c(27, 27, 27, 27),
    y = c(-80, -81, -82, -83),
    z = 0:3,
    timez = Sys.time() + 60 * 60 * (1:4)
  )
  my_sftrack <- suppressMessages(as_sftrack(
    data = df1, group = c("id", "month"),
    time = "timez", coords = c("x", "y", "z")
  ))
  # my_sftrack
  # sftrack -> sftraj -> sftrack
  new_sftraj <- as_sftraj(my_sftrack)
  # new_sftraj
  conv_back_sftrack <- as_sftrack(new_sftraj)
  expect_equal(my_sftrack, conv_back_sftrack)

  # check coordinate dimensions get carried over
  my_sftrack_2d <- suppressMessages(suppressMessages(as_sftrack(
    data = df1, group = c("id", "month"),
    time = "timez", coords = c("x", "y")
  )))
  new_sftraj <- as_sftraj(my_sftrack_2d)
  conv_back_sftrack_2d <- as_sftrack(new_sftraj)
  expect_equal(class(conv_back_sftrack_2d$geometry[[1]])[1], "XY")
})


# test_that("from ltraj conversions", {
#   # multiple bursts become combined
#   df1 <- data.frame(
#     id = c(1, 1, 1, 1),
#     month = c(1,1,2,2),
#     year = c(2020,2020,2020,2020),
#     x = c(27, 27, 27, 27),
#     y = c(-80,-81,-82,-83),
#     z = 0:3,
#     timez = Sys.time() + 60*60*(1:4)
#   )
#
#   ltraj_df <- as.ltraj(xy=df1[,c('x','y')], date = df1$timez,
#    id = df1$id, typeII = TRUE,
#    infolocs = df1[,3:4] )
#   as_sftrack(data = ltraj_df)
#   my_sftrack <- suppressMessages(as_sftrack(data = df1,burst=list(id=df1$id, month = df1$month),
#     time = df1$timez, active_burst = c('id','month'), coords = c('x','y','z')))
#   my_sftrack
#   # sftrack -> sftraj -> sftrack
#   new_sftraj <- as_sftraj(my_sftrack)
#   new_sftraj
#   conv_back_sftrack <- as_sftrack(new_sftraj)
#   all.equal(my_sftrack,conv_back_sftrack)
# })

#' Movements of two raccoons in an urban park in Florida
#'
#' A dataset of two raccoons collared with GPS collars for one month in January 2019
#' in Tree Tops Park, Broward County, Florida, US.
#'
#' @format A data frame with 445 rows and 10 variables:
#' \describe{
#'   \item{animal_id}{ID of individual. TTP: tree tops park, i.e the tagging site.}
#'   \item{timestamp}{The date and time of gps fix in UTC}
#'   \item{latitude}{Latitude in degrees}
#'   \item{longitude}{Longitude in degrees}
#'   \item{height}{Altitude in meters based on satellite positios}
#'   \item{hdop}{Horizontal precision}
#'   \item{vdop}{Vertical precision}
#'   \item{fix}{The number of satellite fixes}
#' }

"raccoon"

###################
# Misc utility functions
###################
#' @title Return a list of sf_POINTS or a data.frame from a sftraj object
#' @name traj_geom
#' @param traj a trajectory geometery from sf_traj
#' @param sfc TRUE/FALSE should the return by an sfc or a list of points. Defaults to FALSE
#' @examples
#' #'
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
#' # Input is a data.frame
#' my_traj <- as_sftraj(raccoon,
#'   time = "timestamp",
#'   error = NA, coords = c("longitude", "latitude"),
#'   group = burstz
#' )
#' print(my_traj, 5, 10)
#'
#' # extract a list of points
#' pts_traj(my_traj)[1:10]
#'
#' # or a data.frame of points
#' coord_traj(my_traj)[1:10]
#' @export
pts_traj <- function(traj, sfc = FALSE) {
  pts <- st_geometry(traj)

  ret <- lapply(pts, function(x) {
    if (inherits(x, "LINESTRING")) {
      st_point(unclass(x)[1L, , drop = TRUE])
    } else {
      x
    }
  })
  if (!sfc) {
    return(ret)
  }

  st_sfc(ret,
    crs = st_crs(pts),
    precision = st_precision(pts)
  )
}

#' @rdname traj_geom
#' @export
coord_traj <- function(traj) {
  # traj = my_sftraj
  pts <- st_geometry(traj)

  if ("XY" %in% class(pts[[1]])) {
    dim <- 2
  } else {
    dim <- 3
  }
  this_seq <- seq(1, dim * 2, by = 2)
  ret <- lapply(pts, function(x) {
    #  x = pts[[499]]
    if (inherits(x, "POINT")) {
      st_coordinates(x)
      x[1:dim]
    } else {
      x[this_seq]
      # st_coordinates(x)[pos, dim]
    }
  })
  do.call(rbind, ret)
}

#' @export
st_coordinates.sftraj <- function(x, return = "all", ...) {
  # x = my_sftraj
  pts <- sf::st_geometry(x)

  ret <- lapply(
    pts,
    function(x) {
      coords <- data.frame(st_coordinates(x))
      coords$Point <- seq_len(nrow(coords))
      coords[, c("X", "Y", "Point")]
    }
  )
  ret <- do.call(rbind, ret)
  choice <- switch(return, start = 1, end = 2, all = c(1, 2))
  ret[ret$Point %in% unlist(choice), ]
}


#' @title Is a trajectory geometry a linestring or a point
#' @description A step is a movement from one point to the next, with an sftraj object
#' this manifests as a linestring. If, however, one of these two points is missing, the sftraj
#' is created as a geometery collection of two points, the beginning and the end point, where one
#' of the steps is NA. This function checks a trajectory geometry if its a linestring and returns
#' a vector of T/F.
#' @export
#' @param x an sftraj object
is_linestring <- function(x) {
  if (inherits(x, "sftraj")) {
    pts <- x[[attr(x, "sf_column")]]
  }
  if (inherits(x, "sfc")) {
    pts <- x
  }
  # if ('XY' %in% class(pts[[1]])) {
  #   dim = c('X', 'Y')
  # } else{
  #   dim = c('X', 'Y', 'Z')
  # }
  vapply(pts, function(y) {
    inherits(y, "LINESTRING")
  }, NA)

  # the sf version might be faster?
  # st_is(x,'LINESTRING')
}

#' @title Summarize sftrack objects
#' @param x an sftrack object
#' @export
summary_sftrack <- function(x) {
  track_class <- class(x)[1]
  # x = my_sftrack
  time_col <- attr(x, "time_col")
  error_col <- attr(x, "error_col")
  sf_col <- attr(x, "sf_column")
  group_col <- attr(x, "group_col")
  sub <- x[, ]
  levelz <- group_labels(x[[group_col]])
  statz <-
    tapply(sub[[time_col]], levelz, function(x) {
      list(
        "begin" = min(x),
        "end" = max(x),
        "points" = length(x),
        "NAs" = sum(is.na(x))
      )
    })

  if (!track_class %in% c("sftrack", "sftraj")) {
    stop(paste0("input class not sftrack or sftraj: ", track_class))
  }
  if (track_class == "sftrack") {
    my_crs <- attr(sub[[sf_col]], "crs")
    lenz <- tapply(sub[[sf_col]], levelz, function(pts) {
      new_pts <- pts[!vapply(pts, sf::st_is_empty, NA)]
      my_sfc <-
        st_sfc(st_linestring(st_coordinates(new_pts)), crs = my_crs)
      st_length(my_sfc)
    })
  }
  if (track_class == "sftraj") {
    lenz <- tapply(sub[[sf_col]], levelz, function(pts) {
      sum(st_length(pts))
    })
  }

  points <- vapply(
    statz,
    FUN = function(x) {
      x$points
    },
    numeric(1)
  )
  NAs <- vapply(
    statz,
    FUN = function(x) {
      x$NAs
    },
    numeric(1)
  )
  begin_time <- lapply(statz, function(x) {
    x$begin
  })
  end_time <- lapply(statz, function(x) {
    x$end
  })
  class(begin_time) <- class(end_time) <- c("POSIXct", "POSIXt")
  attr(begin_time, "tzone") <- attr(x[[attr(x, "time_col")]], "tzone")
  attr(end_time, "tzone") <- attr(x[[attr(x, "time_col")]], "tzone")
  data.frame(
    group = levels(levelz),
    points,
    NAs,
    begin_time,
    end_time,
    length_m = lenz,
    row.names = NULL
  )
}

# recalculates empty geometries (take from sf as it is an internal as well)
sfg_is_empty <- function(x) {
  switch(
    class(x)[2],
    POINT = any(!is.finite(x)),
    MULTIPOINT = ,
    LINESTRING = ,
    CIRCULARSTRING = ,
    CURVE = nrow(x) == 0,
    length(x) == 0
  )
}


#' @title Which grouping/time stamp combos are duplicated.
#' @description This function returns a data.frame of which rows are duplicated and their time stamps.
#' @export
#' @param data a data.frame containing burst or time data (if necessary)
#' @param group a list where each entry is a vector of groupings where length == nrow(data)|nrow(time). Or a character vector describing the column name they are located in data
#' @param time a vector of as.POSIXct time, or a character of the column name where it can be found in data

which_duplicated <- function(data = data.frame(), group, time) {
  # coords = c('longitude','latitude')
  # burst = c(id = 'animal_id', month = 'month')
  # time = 'time'
  # data$time[1] <- data$time[2]
  # data$time[4] <- data$time[5]
  if (length(group) == 1) {
    names(group) <- "id"
  }
  if (all(sapply(group, length) == nrow(data))) {
    # check id in burst
    check_group_id(group)
    group_list <- group
  } else {
    # check names exist
    check_names_exist(data, group)
    # check id in burst
    # check id in burst
    check_group_id(group)
    # create burst list from names
    group_list <- lapply(data[, group, FALSE], function(x) {
      x
    })
    if (!is.null(names(group))) {
      names(group_list) <-
        names(group)
    } else {
      names(group_list) <- group
    }
  }

  if (length(time) == nrow(data)) {
    reloc_time <- time
  } else {
    check_names_exist(data, time)
    reloc_time <- data[[time]]
  }
  check_time(reloc_time)
  group <-
    make_c_grouping(x = group_list)
  gl <- group_labels(group)
  results <-
    unlist(tapply(reloc_time, gl, duplicated))

  rowz <- which(gl[results] == gl & reloc_time[results] == reloc_time)
  data.frame(group = gl[rowz], time = reloc_time[rowz], which_row = rowz)
}

# Get the position of x2, given the time
get_x2 <- function(time) {
  or <- order(time)
  seq_along(time)[or][or + 1][order(or)]
}

#' @title Merge connected lines and create an sf object
#' @description This function returns a sf object grouped by each burst with a geometry column of multilinestrings for each grouping
#' @export
#' @importFrom stats aggregate
#' @param x an sftraj object
merge_traj <- function(x) {
  group_col <- attr(x, "group_col")
  x <- x[order(x[[attr(x, "time_col")]]), ]
  crs <- st_crs(x)
  ret <- stats::aggregate(st_geometry(x), list(group = group_labels(x)), function(y) {
    # y = st_geometry(x)[burst_labels(x, factor = TRUE)=='TTP-041_s']
    geom <- y[st_is(y, "LINESTRING")]
    if (length(geom) > 1) {
      st_line_merge(st_combine(geom))
    } else {
      st_multilinestring(list(st_linestring()))
    }
  })
  ret$geometry <- st_sfc(ret$geometry, crs = crs)
  st_sf(ret, sf_column_name = "geometry")
}

get_point <- function(x, position = "x1") {
  # position = 'x2'
  x <- st_geometry(x)
  if (inherits(x[[1]], "XY")) {
    vapply(
      x, function(y) {
        # y = st_geometry(x)[[10]]
        if (inherits(y, "POINT")) {
          # 3 just represents a non-position here, as NA would fail for empty points
          pos <- switch(position, x1 = 1, x2 = 3, y1 = 2, y2 = 3, xy1 = c(1, 2), xy2 = c(3, 3))
        } else {
          pos <- switch(position, x1 = 1, x2 = 2, y1 = 3, y2 = 4, xy1 = c(1, 3), xy2 = c(2, 4))
        }
        y[pos]
      },
      numeric(1)
    )
  } else {
    vapply(
      x, function(y) {
        if (inherits(y, "POINT")) {
          # 4 just represents a non-position here, as NA would fail for empty points
          pos <- switch(position, x1 = 1, x2 = 4, y1 = 2, y2 = 4, z1 = 3, z2 = 4, xy1 = c(1, 2), xy2 = c(4, 4), xyz1 = c(1, 2, 3), xyz2 = c(4, 4, 4))
        } else {
          pos <- switch(position, x1 = 1, x2 = 2, y1 = 3, y2 = 4, z1 = 5, z2 = 6, xy1 = c(1, 3), xy2 = c(2, 4), xyz1 = c(1, 3, 5), xyz2 = c(2, 4, 6))
        }
        y[pos][[1]]
      },
      numeric(1)
    )
  }
}

#' @title Checks if grouping is ordered by time and then outputs the correct order
#' @export
#' @param group a c_grouping
#' @param time_data a vector of time
#' @param return T/F return the new order or just run check?
check_ordered <- function(group, time_data, return = TRUE) {
  idz <- group_labels(group)

  # may not be as fast as something involving order(time_data, idz)
  isOrdered <-
    all(tapply(time_data, idz, function(x) {
      identical(order(x), seq_len(length(
        x
      )))
    }))
  if (isOrdered & return) {
    return(seq_along(time_data))
  }
  if (!isOrdered) {
    message(paste0("time was not ordered for each burst, reordering output..."))
    if (return) {
      return(order(idz, time_data))
    }
  }
}

#' @title Check if a set of column names are found in a data frame and return an error if not
#' @export
#' @param data a data.frame to check names against
#' @param names the inputted column names
check_names_exist <- function(data, names) {
  # check burst
  if (!is.character(names)) {
    stop("Column names must be a character")
  }
  col_names <- colnames(data)
  test <- !(names %in% colnames(data))
  if (any(test)) {
    stop(paste0(
      paste0(names[test], collapse = " & "),
      " not found in data frame",
      collapse = " "
    ))
  }
}

##### Grouping related checks
#' @title Check there are no NAs in burst
#' @export
#' @param x a c_grouping
check_NA_group <- function(x) {
  group <- x
  if (inherits(x, c("sftrack", "sftraj"))) {
    group <- x[[attr(x, "group_col")]]
  }

  if (any(vapply(group, function(x) any(is.na(x)), logical(1)))) {
    stop("NAs not allowed in groups")
  }
}

#' @title Check there is a grouping id present
#' @export
#' @param x a c_grouping
check_group_id <- function(x) {
  if (!("id" %in% x |
    "id" %in% names(x))) {
    stop("There is no `id` column in group names")
  }
}

# more than one relocation for a burst
check_two_groups <- function(x, ..., active_group) {
  group <- x
  if (inherits(group, "c_grouping")) {
    lvlz <- group_labels(group)
    active_group <- attr(group, "active_group")
  } else {
    lvlz <-
      vapply(group, function(y) {
        paste0(y[active_group], collapse = "_")
      }, NA_character_)
  }
  count <- table(lvlz)
  if (any(count == 1)) {
    message(paste0(paste0(names(count)[count == 1], collapse = " & "), " has only one relocation"))
  }
}
#' @title Are group names equivalent for each s_group?
#' @export
#' @param x a c_grouping
check_group_names <- function(x) {
  group <- x
  if (inherits(x, c("sftrack", "sftraj"))) {
    group <- x[[attr(x, "group_col")]]
  }
  if (length(unique(vapply(group, function(y) {
    paste(names(y), collapse = "")
  }, NA_character_))) != 1) {
    stop("Group names do not match")
  }

  if (any(unlist(lapply(group, function(x) {
    duplicated(names(
      x
    ))
  })))) {
    stop("group names can not be duplicated")
  }
}

# How many bursts are there when combining two different c_groupings
check_active_group <-
  function(x,
           active_group = NULL,
           check_all = T) {
    if (is.null(active_group)) {
      active_group <- active_group(x)
    }
    if (!check_all) {
      check <- all(active_group %in% names(x[[1]]))
    } else {
      check <-
        all(vapply(x, function(y) {
          all(active_group %in% names(y))
        }, NA))
    }
    if (!check) {
      stop("not all new active group names found in current group names")
    }
  }


###################
# coordinate related checks

#' @title fix 0's to NAs in latitude and longitude
#' @param xyz a data.frame of xy or xyz coordinates
#' @return returns a data.frame with 0s replaced with NAs
#' @export
fix_zero <- function(xyz) {
  zero_row <- apply(!is.na(xyz[, 1:2]) & xyz[, 1:2] == 0, 1, any)
  xyz[zero_row, ] <- NA
  return(xyz)
}


#' @title Check if coordinates contain NAs in some columns but not others
#' @param xyz a data.frame of xy or xyz coordinates
#' @export
check_NA_coords <- function(xyz) {
  check_row <- lapply(xyz, function(x) {
    which(is.na(x))
  })
  NAs <-
    vapply(check_row, function(x) {
      any(!unique(unlist(check_row)) %in% x)
    }, NA)
  if (any(NAs)) {
    stop(paste0(
      paste0(c("x", "y", "z")[!NAs], collapse = " "),
      " column has NAs that are not found in other coordinate columns"
    ))
  }
}

# Checks if z coordinates and returns a message
check_z_coords <- function(x) {
  if ("XYZ" %in% class(st_geometry(x)[[1]])) {
    message(
      "Z coordinates found. Note that the vast majority of rgdal calculations are done using 2D geometry"
    )
  }
}
# check_z_coords(my_track)

################
# Time
#' @title Check if time is integer or posix
#' @param time a vector of time
#' @export
check_time <- function(time) {
  # This function was originally envisioned to contain all time checks
  # Currently its not but can expand if we feel its necessary to have
  # is integer or posixct
  if (inherits(time, c("sftrack", "sftraj"))) {
    time <- time[[attr(time, "time_col")]]
  }
  if (!(inherits(time, "integer") | inherits(time, "POSIXct") | inherits(time, "numeric"))) {
    stop("Time needs to be an integer or POSIXct")
  }
}

#' @title Check if time is regular for each burst and returns logical for each burst
#' @param x an sftrack/sftraj object
#' @export
check_t_regular <- function(x) {
  # is complete
  time_col <- attr(x, "time_col")
  idz <- group_labels(x[[attr(x, "group_col")]])
  sftrack <-
    x[check_ordered(idz, x[[time_col]]), ]
  ans <-
    tapply(sftrack[[time_col]], idz, function(date) {
      x1 <- unclass(date[-1])
      x2 <- unclass(date[-length(date)])
      abs(mean(c(x1 - x2)) - (x1[1] - x2[1])) <= 1e-07
    })
  return(ans)
}

#' @title check that time is unique
#' @param x An sftrack/sftraj object or a multi_burst
#' @param time vector of time, not required if given a sftrack object.
#' @export
dup_timestamp <- function(x, time) {
  if (inherits(x, c("sftrack", "sftraj"))) {
    group <- x[[attr(x, "group_col")]]
    time <- x[[attr(x, "time_col")]]
  } else {
    group <- x
  }

  test <-
    tapply(time, group_labels(group), function(y) {
      any(duplicated(y))
    })
  if (any(test)) {
    stop(paste0(
      "groups: ",
      paste0(names(test)[test], collapse = ", "),
      " have duplicated time stamps"
    ))
  }
}

unique_active_group <-
  function(x) {
    # burst = list(burst1,burst2)
    # burst = list(x[[1]],value[[1]])
    if (length(unique(vapply(x, function(y) {
      paste(attr(y, "active_group"), collapse = "")
    }, NA_character_))) != 1) {
      stop("There are more than one possible active groups")
    }
  }

#' @title Methods for plotting sftrack/sftraj
#' @name plot_sftrack
#' @export
#' @param x 'sftrack' or 'sftraj' object
#' @param y ignored
#' @param key.pos Integer; side to plot a color key: 1 bottom, 2 left,
#'     3 top, 4 right; set to NULL to omit key, or -1 to select
#'     automatically (defaults to 4; see \code{\link[sf]{plot_sf}} for
#'     more details).
#' @param key.width Amount of space reserved for the key, including
#'     labels (see \code{\link[sf]{plot_sf}} for more details).)
#' @param step_mode Logical; whether to plot in step mode, see
#'     details, defaults to TRUE, unless there are more than 10,000
#'     steps.
#' @param ... Further arguments passed to 'plot.sf'. Among others,
#'     arguments for the key are set differently in 'sftrack' to allow
#'     for longer labels by default (but can be nevertheless
#'     adjusted).
#' @details Step mode refers to considering the trajectory as
#'     individual 'steps', in the case of plot this means it will plot
#'     each line & point individually. This approach is much slower to
#'     plot with large objects, and is thus turned off when
#'     n(steps)>10,000.  The alternative, much faster method is to
#'     merge the steps into a multilinestring as continuous lines.
#' @method plot sftrack
#' @importFrom graphics plot lcm
#' @examples
#'
#' ## Prepare an 'sftrack' object:
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- c(id = "animal_id")
#' my_sftrack <- as_sftrack(raccoon,
#'   time = "timestamp",
#'   coords = c("longitude", "latitude"),
#'   group = burstz
#' )
#'
#' ## Plotting with sftrack is just like sf. `...` will accept most
#' ## arguments as 'plot.sf':
#' plot(my_sftrack, axes = TRUE, lwd = 5, cex = 5, bgc = "gray50")
#'
#' ## sftraj will as well for the most part; however it is a more complex
#' ## structure that combines points and steps (in step mode):
#' my_sftraj <- as_sftraj(raccoon,
#'   time = "timestamp",
#'   coords = c("longitude", "latitude"),
#'   group = burstz
#' )
#' plot(my_sftraj, lwd = 5, cex = 5, bgc = "gray50", graticule = TRUE)
#'
plot.sftrack <- function(x, y, key.pos, key.width, ...) {
  if (missing(key.pos))
    key.pos <- 4
  if (missing(key.width))
  {
    if (key.pos %in% c(2, 4))
      key.width <- lcm(4)
    else key.width <- lcm(1.8)
  }
  group_col <- attr(x, "group_col")
  x[[group_col]] <- group_labels(x)
  class(x) <- setdiff(class(x), c("sftrack"))
  x <- x[group_col]
  plot(x, reset = FALSE, key.pos = key.pos, key.width = key.width, ...)
}

#' @title Methods for plotting sftrack/sftraj
#' @export
#' @rdname plot_sftrack
#' @method plot sftraj
#' @importFrom graphics plot
plot.sftraj <- function(x, y, key.pos, key.width, ..., step_mode) {
  if (missing(key.pos))
    key.pos <- 4
  if (missing(key.width))
  {
    if (key.pos %in% c(2, 4))
      key.width <- lcm(4)
    else key.width <- lcm(1.8)
  }
  if (missing(step_mode)) {
    step_mode <- if (nrow(x) < 10000) {
      TRUE
    } else {
      FALSE
    }
  }
  group_col <- attr(x, "group_col")
  if (step_mode)
  {
    x[[group_col]] <- group_labels(x)
    class(x) <- setdiff(class(x), c("sftraj"))
    x <- x[group_col]
  } else
  {
    warning("Step mode disabled. See details in the help of 'plot.sftraj'.")
    x <- merge_traj(x)
  }
  plot(x, reset = FALSE, key.pos = key.pos, key.width = key.width, ...)
}

#' @title Function to plot sftrack objects in ggplot
#' @name geom_sftrack
#' @description
#' This function can be added to ggplot() to plot an sftrack and sftraj
#' Function does not yet work with ggplot grammer so you must but data= in this function
#' @details step mode refers to considering the trajectory as individual 'steps', in the case of plot this means it will
#' plot each line & point individually. This approach is much slower to plot when n(steps)>10,000.
#' The alternative method is to merge the steps into a multilinestring of continuous lines. This is much faster to plot.
#' @param mapping mapping aesthetics for ggplot.
#' @param data the sftraj or sftrack object.
#' @param ... arguments to passed to ggplot
#' @param step_mode TRUE/FALSE, whether to plot in step_mode, See details
#' @examples
#' #'
#' require("ggplot2")
#' data("raccoon")
#' raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
#' burstz <- c(id = "animal_id")
#'
#' # sftraj will as well for the most part, however as its a more complex
#' # structure to speed up plotting.
#' my_sftraj <- as_sftraj(raccoon,
#'   time = "timestamp",
#'   coords = c("longitude", "latitude"),
#'   group = burstz
#' )
#'
#' ggplot() +
#'   geom_sftrack(data = my_sftraj)
#' @export
geom_sftrack <- function(mapping, data, ...) {
  UseMethod("geom_sftrack")
}

#' @rdname geom_sftrack
#' @export
geom_sftrack.sftrack <-
  function(mapping = ggplot2::aes(),
           data = NULL,
           ...) {
    group <- group_labels(data)
    class(data) <- class(data)[2:3]
    list(
      ggplot2::geom_sf(data = data, ggplot2::aes(color = group), ...)
    )
  }

#' @name geom_sftrack
#' @export
geom_sftrack.sftraj <-
  function(mapping = ggplot2::aes(),
           data = NULL,
           ..., step_mode = FALSE) {
    # data = my_sftraj
    group_col <- attr(data, "group_col")
    if (step_mode) {
      data[[group_col]] <- group_labels(data)
      class(data) <- setdiff(class(data), c("sftraj"))
      data <- data[group_col]
    } else {
      data <- merge_traj(data)
    }
    group <- data[[group_col]]
    list(
      ggplot2::geom_sf(data = data, ggplot2::aes(color = group), ...)
    )
  }

# ggplot() + geom_sf(data = data, aes(color = .data[[group_col]]))

---
title: "3. Structure of sftrack/sftraj objects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{3. Structure of sftrack/sftraj objects}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)
library("lwgeom")

```

This is section will go more depth into the structure of sftrack/sftraj objects and how to work with them.

To begin, `sftrack` and `sftraj` objects are essentially data.frame objects with the 3 required columns (group, geometry, and time). However, they are also of the subclass `sf`. This allows them to act like sf objects when working with functions in the `sf` package but act as `sftrack` objects for all other actions. It should be noted that when possible sftrack objects will mimic `sf` functionality and thus many of the same words and tactics are used.

```{r}
library("sftrack")

# Make tracks from raw data
data("raccoon", package = "sftrack")
#raccoon <- read.csv(system.file("extdata/raccoon_data.csv", package="sftrack"))
raccoon$month <- as.POSIXlt(raccoon$timestamp)$mon + 1

raccoon$time <- as.POSIXct(raccoon$timestamp, tz = "EST")
coords <- c("longitude","latitude")
group <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon+1)
time <- "time"
error <- "fix"
crs <- 4326
# create a sftrack object
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error, crs = crs)

# create a sftraj object
my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time, error = error, crs = crs)

```

In order for `sftrack` to function as an `sf` object, we create the data object as an `sf` object first (using st_as_sf()), and then add the `sftrack` attributes to the object. The class of an sftrack object is `sftrack` -> `sf` -> `data.frame` although the data.frame class is rarely called upon.

There are five attributes total to an `sftrack` object, two of these are created by `sf` (`sf_column` and `agr`), and the additional three are created by `sftrack` (`group_col`, `time_col`, and `error_col`).

```{r}
attributes(my_sftrack)[-(1:2)]
```

The `sftrack` level attributes are simply pointers to the data. Any attributes relevant to the grouping or geometry are stored in those columns themselves. 

## Geometry

The geometry column is an `sfc` object and contains the important spatial information for the track. As NA points are common and important in movement data, we create the `sfc` object with the option `na.fail = TRUE`. 
```{r}
my_sftrack$geometry
```

The `sfc` column varies in structure dependent on the movement class. An `sftrack` is a collection of `POINTs` while an `sftraj` is a `GEOMETRY` collection of `POINTs` and `LINESTRINGs`. 


```{r}

df1 <- data.frame(
    id = c(1, 1, 1, 1,1,1),
    month = c(1,1,1,1,1,1),
    x = c(27, 27, 27, NA,29,30),
    y = c(-80,-81,-82,NA, 83,83),
    timez = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60*60*(1:6)
)

test_sftraj <- as_sftraj(data = df1, group = list(id = df1$id, month = df1$month),
    time = df1$timez, active_group = c("id","month"), coords = df1[,c("x","y")])
test_sftraj$geometry

```


## Grouping

The grouping column is very specialized, and we will cover it in its own section. To begin, the novel attributes it stores are the `active_group` and the `sort_index` which is a factor of the current active groups. The grouping class consists of single row level group: `s_groups` (a.k.a single groups) and a column level collection of s_groups called a `c_grouping` (a.k.a column/collection grouping). This column acts as a robust storage column for the groupings and is maintained across a `sftrack` object.

```{r}
attributes(my_sftrack$sft_group[1:10])
summary(my_sftrack)
```


## Time

The time column must be either an integer or POSIXct and the column must be of one type of time. Beyond that there is not much specialized functionality in the column. Sftrack uses the time column to order outputs for analysis, and attempts to order outputs when originally making an sftrack object, however, the data.frame is not required to be ordered for analysis. A call to `check_ordered()` is called before analysis, and otherwise it is assumed the order does not matter. This is particularly true for a sftraj, where the geometry level contains information about t1 and t2. 


## Error

The error column is the column with the relevant error information for the spatial points in it. At present we have not built particular functionality but plan to in the future or reserve this for other developers to build upon.


## Subsetting

An sftrack object acts like a data.frame and sf whenever appropriate. Because of this you can subset an sftrack object as you would a data.frame. 

In this way row subsetting is very straight forward, as each row represents an individual point in time.

```{r}

my_sftrack[1:10,]
```

Unlike a data.frame, however, sftrack attempts to retain the geometry, group, and time columns, in order to maintain sftrack status. This is similar to how `sf` deals with the geometry column.

```{r}

my_sftrack[1:3,c(1:3)]
```

To turn off this feature, use the `drop = TRUE` argument which returns a
data.frame object instead. If you would like to revert to an sf object, `sf::st_sf(data)` will return the object to an `sf` object.

```{r}
my_sftrack[1:3,c(1:3), drop = TRUE]
```

`sftraj`s work nearly the same as `sftrack`s, however because they are a step model where the steps are modeled as step1 (t1 ->t2) its important to note that subsetting will not automatically recalculate any new steps for you even if the original t2 point has been deleted.

If your subsetting will also change the end points for steps, then you can recalculate using `step_recalc()`. The output which is your original sftraj object but with the geometry column recalculated to the new t2s based on the timestamp. The results of which can be wildly different than the original subsetted data.frame. So be careful.

```{r}
plot(my_sftraj, main = "Original")
new_traj <- my_sftraj[seq(10, nrow(my_sftraj), 10), ]

plot(new_traj, main = "Before recalculation")

plot(step_recalc(new_traj),  main = "After recalculation")
```


## Some basic methods and functions of sftrack and sftraj objects


### print

The `print()` layout is a modified version of the `sf` print function. It returns important info summarazing the sftrack object like the geometry information and burst information. The print function defaults to printing 1000 rows and displaying all the columns. This can be modified using the `n_row` and `n_col` arguments, which subset the printed output in the repsective axis. When using `n_col` the display will show the `grouping` `geometry`, and `time` fields as well as any other columns starting from column 1 until **#columns + 3 = n_col**. `n_col` and `n_row` are optional arguments, and if values are not supplied they default to the data.frame defaults. Note : `n_row` is not a corrolary to `head()`, as `head()` physically subsets the data while the `n_row` option just modifies the printed output. 

If

```{r}

print(my_sftrack, 5, 10)
```


### summary

`summary()` works as youd expect for a data.frame, except it displays the grouping column as a count of each active_group combination and the `active_group `for that column.  
```{r}
summary(my_sftrack)
```


### summary_sftrack

`summary_sftrack()` is a special summary function specific for sftrack objects. It summarizes the data based on the beginning and end of each grouping as well as the total distance of the grouping. This function uses `st_distance` from the `sf` package and therefore outputs in units of the CRS. In this example the distance is in meters.

```{r}
summary_sftrack(my_sftrack)

```

You can also trigger this function by using `summary(data, stats = TRUE)`

```{r}
summary(my_sftrack, stats = TRUE)
```

---
title: "1. Overview of sftrack"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{1. Overview of sftrack}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)

```

## Definitions of movement data

We start by defining the main objects that `sftrack` is dealing
with. At the top level, there is a **path**, i.e. the real-world
(continuous) curve made by moving objects (Fig. 1).

```{r, out.width = "75%", fig.align = "center", fig.cap = "Fig.1: Glossary introducing the main concepts.", echo = FALSE}
knitr::include_graphics("glossary.png")

```

In practice, this curve is sampled (recorded) at discrete times to
collect **locations**, with or without temporal attributes (think of
non-invasive monitoring of animals such as snow tracks vs. direct
observations), i.e. in the form of $(x, y)$ or $(x, y, t)$
coordinates. Taken together, locations form the bulk of **occurrence
datasets**, such as the massive amount of biodiversity data found on
[Global Biodiversity Information Facility
(GBIF)](https://www.gbif.org/).

If the organism monitored is known and identified ($id$ attribute), we
can collect **tracking locations** (Fig. 1), i.e. $(x, y, t, id)$
coordinates. An ordered series of tracking locations together are in
turn the vertices of a **track**. A track is thus composed of a series
of locations, i.e. the "complete spatio-temporal record of a followed
organism, from the beginning to the end of observations"[^turchin].
Note that $id$ can be further refined as a general grouping factor
(e.g. collar, season, behavior, etc.), which basically split tracking
data into sub-tracks.

[^turchin]: Turchin, P. (1998). Quantitative analysis of movement: measuring and modeling population redistribution in animals and plants. Sinauer Associates, Sunderland, MA, USA.

Finally, tracking data can be modeled considering the temporal
autocorrelation between successive locations as part of the movement
information[^martin]. The most common approach for that purpose is
based on a model of **steps** (Fig. 1), defined as straight-line
segments between two successive tracking locations. Treating movement
steps as straight lines is of course an idealization, and the most
parsimonious approach (other complex approaches exist in some fields,
such as splines and Bezier curves), but at sufficient resolution, no
significant information is lost. A sequence of steps from the same
individual (not necessarily connected) finally forms a **trajectory**.

[^martin]: Martin, J., Tolon, V., Van Moorter, B., Basille, M., & Calenge, C. (2009). On the use of telemetry in habitat selection studies. In D. Barculo, & J. Daniels (Eds.), Telemetry: Research, Technology and Applications (pp. 37–55). Nova Science Publishers Inc.


### Tracking and movement data analyses

Because of the dual nature of tracking and movement data, each data
type allows different analyses based on the underlying geometric
representation (points vs. steps). The following table summarizes in
which case the use of tracking vs. movement data is appropriate:

| Data                     | Tracking data                | Movement data                                       |
|:-------------------------|:----------------------------:|:---------------------------------------------------:|
| Geometric representation | Points                       | Steps                                               |
| Applications             | Home ranges                  | Random walks                                        |
|                          | Resource selection functions | Step selection functions                            |
|                          | K-select                     | typical Hidden-Markov models and State-space models |
|                          |                              | typical segmentation methods                        |
|                          |                              | First passage time and Residence time               |
Table: Table 1: Overview of applications based on tracking and movement data.

While this is out of the scope of the `sftrack` package to provide
methods for these space use analyses, the package however provide the
underlying infrastructure for each case, and a few tools to handle
movement of both living organisms and inanimate objects.


## A conceptual model for movement

To organize these elements of data and how they relate to one another
and to properties of the real world entities (here a moving object),
we followed the methodology for conceptual modeling of geographic
information that is specified in the International Standards (ISO/TC
211), which relies on a standard conceptual model specified in Unified
Modeling Language (UML)[^roswell].

[^roswell]: Roswell, C. (2011). Modeling of geographic information. Springer Handbook of Geographic Information, 3–6. http://dx.doi.org/10.1007/978-3-540-72680-7_1


### Tracking locations

The basic element of the movement data is a **tracking location**. A
tracking location is, in its simplest form, a point defined by its
spatial coordinates $(x, y, z, t)$, with $z$ being optional,
associated to the corresponding timestamp $t$ (note that the time can
also be a simple integer defining the order, for instance in the case
of a track in the snow), and the identity of the moving object. This
is the raw data of movement, which is usually directly provided by a
tracking device.

![Fig. 2: The location model.](conceptual-model-loc.png "Fig. 2: The location model.")

Additionally, "grouping" information can be provided for the location:
this needs to be *a minima* the individual (`ind`), e.g. the actual
moving object (for instance an animal), but can also be extended to
other hierarchical level of the data, such as the year, the grouping, the
movement behavior, etc.

Spatial points are never exact and can be characterized by an
error. Typically, GPS and Argos data will provide information on the
quality of the location, such as satellite data or dilution of
precision (for GPS data). The challenge with this information on the
error is the complete lack of standardization — as such, this will be
left open for the time being.

Finally, a tracking location can also include any information at the
level of the point, such as environmental data measured at the
location (e.g. land cover).


### Tracking data

Tracking data, in the form of **tracks**, is a series of **tracking
locations**. This can be modeled as a collection of tracking
locations, unique and ordered in time by individual. As such,
locations are the vertices of a single track object; conversely, all
locations can be directly extracted from a single track (without any
loss of information).  The strong constraint on time is necessary and
sufficient to organize the data in a track.

![Fig. 3: The track model.](conceptual-model-track.png "Fig. 3: The track model.")


### Steps

A **step** is specifically defined as the straight-line segment
connecting two successive **tracking locations**. In other words, it
takes two and only two locations to define a step, with the
constraints that they belong to the same grouping level (individual in
its simplest form), and that the second location comes temporally
after the first one.

![Fig. 4: The step model.](conceptual-model-step.png "Fig. 4: The step model.")

Note that a step can also include information at the level of the line
segment. This can be for instance the amount of forest along the step,
or the movement rate (i.e. speed).


### Trajectories

Similarly to tracks, **trajectories** are defined by a series of
**steps**.  A trajectory is thus modeled as a collection of steps,
unique and ordered in time by individual. As such, steps are the
vertices of a single trajectory object; conversely, all steps can be
directly extracted from a single trajectory (without any loss of
information).  The strong constraint on time is necessary and
sufficient to organize the data in a trajectory.

![Fig. 5: The trajectory model.](conceptual-model-traj.png "Fig. 5: The trajectory model.")

Given the double equivalence between tracking locations and tracks, on
the one hand, and steps and trajectories, on the other hand, it
follows that there is also a direct equivalence between a track and a
trajectory (based on the same movement data); lossless conversion is
possible in both directions.

![Fig. 6: Full model.](conceptual-model-full.png "Fig. 6: Full model.")


## Implementation in the `sftrack` package

The `sftrack` package deals with all elements presented above,
specifically **tracks** and **trajectories**. We'll begin with a brief
overview of an `sftrack` object. An `sftrack` object is a data object
that describes the movement of a subject, it has $(x,y)$ coordinates
(sometimes $z$), some measurement of time $t$ (clock time or a
sequence of integers), and some grouping variable that identifies the
subjects. For the spatial aspects of `sftrack`, we are using the
package `sf`, a powerful tool that lets us quickly calculate spatial
attributes and plot with ease with its full integration with `rgdal`.

An `sftrack` object has 4 parts to it, 3 of which are required:

 - **Geometry**: This is stored as an `sfc` object from `sf`. Accepts
   $(x,y,z)$ coordinates.
 - **Group**: This is the grouping variables, which contains at
   minimum an `id` field to identify the subject. Grouping variables
   can be turned on at various stages based on the 'active' grouping,
   allowing for the column to be more dynamic than a standard grouping
   column.
 - **Time**: Either a POSIX time object or an integer.
 - **Error**: (optional). This is the field with error data in it. At
   present the column name is being stored as an attribute, but it
   does not have any real functionality yet.

There are two different classes of objects: `sftrack` for **tracks**,
`sftraj` for **trajectories**.


### `sftrack` objects

An `sftrack` object stores tracking data (set of **tracks**) in a
`data.frame`, where the geometries are stored as a `POINT` for each
row.

```{r sftrack-overview, echo = FALSE}
library("sftrack")
data(raccoon)
raccoon$timestamp = as.POSIXct(raccoon$timestamp, tz = "EST5EDT")
my_sftrack <- as_sftrack(
  data = raccoon,
  coords = c("longitude","latitude"),
  time = "timestamp",
  group = "animal_id",
  crs = 4326)
head(my_sftrack)
plot(my_sftrack)

```


### `sftraj` object

An `sftraj` object stores movement data (set of **trajectories**) in a
`data.frame`, where the geometries are stored as a `LINESTRING` for
each row. The linestring is a line built from the two points at 
$t_1 → t_2$. This means that when an `sftraj` object is modified the
steps likely need to be recalculated. Internally, anytime the active
group is changed, then the geometry is recalculated.

```{r sftraj-overview, echo = FALSE}
my_sftraj <- as_sftraj(my_sftrack)
head(my_sftraj)
plot(my_sftraj)

```

---
title: "5. Getting Spatial with sftrack"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{5. Getting Spatial with sftrack}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)
library("lwgeom")

```

```{r}
library("sftrack")
# Make tracks from raw data
# raccoon <- read.csv(system.file('extdata/raccoon_data.csv', package='sftrack'))
data("raccoon", package = "sftrack")
raccoon$month <- as.POSIXlt(raccoon$timestamp)$mon + 1

raccoon$time <- as.POSIXct(raccoon$timestamp, tz = "EST")
coords <- c("longitude","latitude")
group <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon+1)
time <- "time"
error <- "fix"
crs <- 4326
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error, crs = crs)
my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time, error = error, crs = crs)
```


## Geometry column

As stated earlier, the geometry column is built using `sf`, so the column functions exactly as it would in `sf`. You can modify it and redefine it using the `sf` tools. More specifically the geometry column of an sf_track object is an `sfc` column. The main difference between a standard `sf` object created using `st_as_sf` is that we automatically allow empty geometries, where as this option is turned off by default in `st_as_sf()`.

```{r}
my_sftrack$geometry
```

An `sftrack` object is simply an `sfc` of `sfc_POINTS`, this contrasts with an `sftraj` object which is a mixture of a `POINT` and `LINESTRING`. This is because a trajectory can have a start point and an NA end point, a line segment, or an NA and an end point. This allows no-loss conversion back and forth between `sftrack` and an `sftraj`, and because linestrings can not have a NULL point in them.

```{r}
my_sftraj$geometry
```

This does mean that not all `sf` functions will handle an `sftraj` object like it would an `sftrack` if there are NAs in the data set. In these cases `st_is_empty()` can help to subset the points that contain geometry data.


## Working with sf

As an sftrack object is an sf object essentially all of the sf functions will work on it.

```{r}
library("sf")
st_length(my_sftraj)[1:10]

df1 <- data.frame(
  id = c(1, 1, 1, 1),
  time = as.POSIXct("2020-01-01 12:00:00", tz = "UTC") + 60*60*(1:4),
  x = c(1, 3, 3, 2),
  y = c(1, 1, 3, 4)
)

road <- st_linestring(rbind(
  c(1, 2),
  c(5, 2),
  c(5, 0)
)
)

animal1 <- as_sftraj(df1)
plot(animal1)

plot(road,  col = "red", add = TRUE)

# Does the animal cross the road?

any(st_intersects(animal1, road, sparse = FALSE))

# When?
animal1$time[st_intersects(animal1, road, sparse = FALSE)]

# How often does the animal stay near the road?
st_is_within_distance(animal1, road, 1)

# How close is the animal from the road?

st_distance(animal1, road)
```

The only thing to remember, is that a sftraj is a `GEOMETRY` column, and occasionally a function may not work with it. In those cases `is_linestring()` can be used to filter out points that do not have a t2.


## Plotting


### Base plotting 

`sftrack` is built to work with `sf` base plot methods. This means you can use most of the `sf` plot methods, `sftrack` largely just controls the grouping of the plot then feeds it back to `plot.sf()`. 


```{r}
plot(my_sftraj)
plot(my_sftraj, key.pos = 4, key.width = lcm(4), key.length = 1)
get_key_pos(my_sftraj)
```

This means that everytime you change the active_group, the plot view will change.

```{r}
active_group(my_sftraj$sft_group) <- "id"
active_group(my_sftraj$sft_group)
plot(my_sftraj)
```

Most arguments for `plot.sf` are available to use as additional arguments to `plot`. 

```{r}
plot(my_sftrack, axes = TRUE, cex = 5)
```


### ggplot2

This is a work in progress, but theres a geom_sftrack function that feeds `geom_sf` with the correct plotting information. Like `geom_sf` you input `data` into the geom_sftrack function and not into `ggplot()`.  Again ggplot assumes active_group is the grouping variable. Plots vary slightly based on if they're sftrack of sftraj class.

Geom_sftrack is essentially geom_sf(data = data, aes(color = group_labels(data))) with NULL points subetted out. This may help when a user requires more advanced modification than the geom_sftrack allows.

```{r}
library("ggplot2")
ggplot() + geom_sftrack(data = my_sftraj)
```

You can use geom_sftrack just like any other ggplot layer, which means you can continue to make manuscript quality plots.

```{r}

cols <- c("TTP-041_1" = "dodgerblue1", "TTP-041_2" = "darkseagreen2",
          "TTP-058_1" = "darkorchid1", "TTP-058_2" = "khaki3")

ggplot() + geom_sftrack(data = my_sftrack, size = 3, alpha = 0.5) +
scale_color_manual(values = cols) + 
  ggtitle("Raccoons at Tree Tops Park Winter 2020") +
  theme_bw() + 
  theme(axis.text.x = element_text(angle = 45, hjust = 1)) 
```


## Special functions for working with an sftraj

To help with working with more complex sftraj objects, there is a growing suite of `sftraj` specific functions:


### Step mode in plotting

`sftraj` objects create linestrings for each row of data, as each row is not assumed to be related to the next row of data. This data structure may become inefficient to work with when plotting large data sets. When appropriate these trajectories can be generalized by combining linestrings where lines meet.

For plotting purposes we can create these linestrings quickly and plot them at much faster speeds than individual lines. 

For `plot` and `geom_sftrack` there is an argument called **step_mode** that refers to whether youd like to plot the lines individually (`step_mode = TRUE`), or generalize them into connected linestrings (`step_mode = FALSE`). By default step_mode is set to TRUE.

```{r}
library("ggplot2")
ggplot() + geom_sftrack(data = my_sftraj, step_mode = TRUE)
```

You'll notice that the appearance of the plot changes as the POINTs are also displayed. This is because step mode adds POINTS to the plot that contain a fill as well as a color property.


### coord_traj 

This function returns a data.frame (x,y,z) of the point at t1 of each sftraj geometry. It works nearly identically to `sf::st_coordinates()`.

```{r}
coord_traj(my_sftraj$geometry)[1:10, ]
```


### pts_traj 

If youd like to retain the geometries but still pull out t1 point you can use `pts_traj()`. This functions returns a list of the beginning point of each sftraj geometry, or an sfc column when using the argument `sfc = TRUE`.

```{r}
pts_traj(my_sftraj$geometry)[1:5]

pts_traj(my_sftraj$geometry, sfc= TRUE)[1:5]
```

### is_linestring 

May help if you'd like to quickly filter an `sftraj` object to just contain pure linestrings. `is_linestring()` returns TRUE or FALSE if the geometry is a linestring. This does not recalculate anything, it just filters out steps that contained NAs in either phase. Its nearly identical to st_is(x,'LINESTRING'), but may be more intuitive for users.

```{r}
is_linestring(my_sftraj$geometry)[1:10]
new_sftraj <- my_sftraj[is_linestring(my_sftraj$geometry), ]
head(new_sftraj)
```


### Calculating step metrics of an sftraj

For use in movement models, you may need to calculate the dx, dy, length, and turn angles of each step. You can do that in `sftrack` using `step_metrics()`. It should be noted it will accept an `sftrack` object, however, it first converts the geometries internally to `sftraj` geometries and then calculates step metrics. As with other `sf` objects, the return is assumed to be in the units of the `crs` when not specified. Absolute angle is measured in radians.

```{r}
step_metrics(my_sftraj[1:10, ])
```

---
title: "4. Fantastic Groups and how to use them"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{4. Fantastic Groups and how to use them}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)

```


Groupings are a big emphasis in `sftrack`. Structurally they are built with in the same vein as `sfc` and `sfg` classes in `sf`. 

To begin an `s_group` is a singular grouping. Its whats stored at the row level. 
A `c_grouping` is a collection of `s_groups` and exists at the column level. Groups also have an `active_group` argument, which turns on and off certain groups for analysis and plotting purposes.

We start by looking at the structure of a `c_grouping`

```{r}
library("sftrack")
data("raccoon", package = "sftrack")
#raccoon <- read.csv(system.file("extdata/raccoon_data.csv", package="sftrack"))
group_list <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon + 1)

cg1 <- make_c_grouping(x = group_list, active_group = c("id", "month"))
str(cg1)
cg1[[1]]
```

A grouping contains group related information. The `id` of the subject/sensor is the lowest level of grouping for the data. Any additional grouping variables are optional. 

A c_grouping is simply a collection of `s_group`s. The `s_group` is where the grouping data is stored, and can be modified at the row level. The `s_groups` main job is to store row level grouping information and maintain consistency of the grouping variables.


## Single groups (s_group) 
An s_group is the grouping variables for a single row of data.  

You can make an `s_group` object using `make_s_group()`, and giving it a list with the group variables named. In this example we have a single sensor named 'TTP_058' from a raccoon, and an additional grouping variable of month (entered as its numeric interpretation).

All grouping information is converted and stored as a character in the s_group.

```{r}

singlegroup <- make_s_group(list(id = "TTP_058", month = 4))
str(singlegroup)
```

Because `s_group`s are simply lists, you can edit individual elements in an s_group
```{r}
singlegroup 
singlegroup[1] <- "CJ15"
singlegroup$month <- "5"
str(singlegroup)
```


## Column groupings (c_grouping) 

C_groupings are a collection of s_groups with the same grouping names and an 'active_group' which is a subset of all the available group names.

Similarly to s_group you can make a c_grouping with `make_s_group`. The argument `x` takes a list where each element is a vector indicating the named groupings as well as a vector of the active groups.

```{r}
group_list <- list(id = rep(1:2, 10), year = rep(2020, 10))
cg <- make_c_grouping(x = group_list, active_group = c("id", "year"))
str(cg)
```

You can also make a c_grouping by concatenating multiple s_groups. All `s_groups` must have the same names or an error is returned.

```{r}
a <- make_s_group(list(id = 1, year = 2020))
b <- make_s_group(list(id = 1, year = 2021))
c <- make_s_group(list(id = 2, year = 2020))
cg <- c(a, b , c)

summary(cg)
```
You'll notice sftrack warns if any grouping combination only has one relocation. This may be relevant for different kinds of analysis where you need more than one point to consider a movement. 

You can also combine c_grouping together with `c()`. All names and active_groups must be the same.  

```{r}
cg_combine <- c(cg,cg)
summary(cg_combine)
```
You can also edit groupings like a list, but you must replace it with an s_group object.
```{r}
cg[1]
cg[1] <- make_s_group(list(id = 3, year = 2019))
cg[1]

```

And the group names must match the ones in the c_grouping or an error is returned
```{r}
# Try to add an s_group with a month field when the original group had year instead 
try( cg[1] <- make_s_group(list(id = 3, month = 2019)) )
```


## Selecting a grouping

As `c_groupings` are stored as lists, it can be difficult to refer to a single group or combination of groupings. This is where the sort_index can come in handy. The sort_index is a factor of the combined active_group variables for each row where paste(id, name1,name2,...,sep='_')). The sort index is remade every time the active group changes, and therefore can be used to subset. You can access the sort index using `group_labels()`. 

```{r}
group_list <- list(id = rep(1:2, 10), year = rep(2020, 10))
cg <- make_c_grouping(x = group_list, active_group = c("id", "year"))
group_labels(cg)[1:10]

# Subsetting a particular sensor from our raccoon data

data("raccoon", package = "sftrack")
raccoon$month <- as.POSIXlt(raccoon$timestamp)$mon + 1

raccoon$time <- as.POSIXct(raccoon$timestamp, tz = "EST")
coords <- c("longitude","latitude")
group <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon+1)
time <- "time"

my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time)
head(my_sftraj[group_labels(my_sftraj) %in% c("TTP-058_1"), ])

```

You can also subset by entering the group label of the group itself in either the c_grouping or the sftrack/sftraj object:

```{r}
head(cg["1_2020"])

sub <- my_sftraj["TTP-058_1", ]
print(sub, 5, 3)
```

You can refer to the levels of the sort index using `group_names`, this returns levels() of the sort index.

```{r}
group_names(cg)
```


## Active group 

The active_group is a simple yet powerful feature. It dictates how your data is grouped for essentially all calculations. It can also be changed on the fly. You can view and change the active group of a column grouping with `active_group()`. Once changed, it recalculates the sort_index and in some cases recalculates the geometries.

Active groups can be changed for any sftrack/straj/c_grouping

```{r}
# sftrack
active_group(my_sftraj)
summary(my_sftraj, stats = TRUE)

# change the active group to id only
active_group(my_sftraj) <- c("id")
active_group(my_sftraj)

summary(my_sftraj, stats = TRUE)

# column groupings work the same way
active_group(cg)
active_group(cg) <- "id"
active_group(cg)

```

---
title: "2. Reading in an sftrack"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{2. Reading in an sftrack}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    echo = TRUE,
    fig.width = 6,
    fig.asp = 0.7
)

```

Creating sftrack objects is relatively straight forward and can be read in a variety of ways including as a standard `data.frame`, `sf` object, or `ltraj` object (from `adehabitatLT`).  


## Loading in raw data

To create `sftrack` objects data we use the `as_sftrack()` (track) or `as_sftraj()` (trajectory) function, depending on your desired output. Both have the same arguments but differ in the way the geometry field is calculated.  

### Vector vs. data.frame inputs

`as_sftrack()`/`as_sftraj` accepts 2 kinds of raw data for each of the 4 required parts. Either a vector/list representing the data where length = nrow(data), or the column name where the data exists. For any `sftrack` component you can input either vector data or the column name for any variable, and can mix types between arguments.

**Vector** inputs to `as_sftrack` generally involve feeding as_sftrack the data itself where length(vector) == nrow(data). Or a list where each component adheres to this rule. If using entirely vector inputs for grouping, geometry, time, and error then `data` is not required. 

**data.frame** inputs on the other hand are simply character vectors describing the column name in `data` where the information is found.

### Arguments 

**data**: A data.frame containing your data. Data is optional if all inputs are done in vector mode where the appropriate vectors are given for geometry, grouping, and time. If inputting vector mode, sftrack creates column names `sft_group`, `sft_timestamp`, and `sft_error` in data frame. Names can be manually changed using `group_name`, `time_name`, and `error_name` and changing `overwrite_names` to FALSE will write over the data if the column already exists in `data`.

**group** (required): A list with named vectors to group the sftrack. One group must be named `id`, but otherwise can be infinite number of grouping variables. Or a vector naming the column names for each grouping categories. If only one variable is given then that vector is assumed to be the id column.  

**coords** (required): The x,y,z coordinates to calculate geometries via `sf`. Accepts either a vector of `c(x,y,z)` describing which column the coordinates can be found, or a list(x=, y=, z=) with vectors for each coordinate. z is optional. NAs are allowed, alhough NAs must exist through the entire row otherwise an error is thrown. This is the same argument from `sf` functions and feeds into `st_as_sf()` internally.    

**time** (required): Time information in either `POSIXct` or as an `integer`. Accepts either a vector of time, or the column name found in `data`. The outputted object will be sorted by the time column.  

**error**: Error information for the associated xyz point. Accepts either a
vector of the error, or the column name found in `data`. If not given, defaults to `NA`.

**crs**: The coordinate references system/projection of the data, as
implemented by `sf`. It accepts an integer with the EPSG code, or a character with proj4string. If none is supplied `crs` is set as `NA` and can be set later using `sf::st_crs()` from `sf`.

**active_group** (required): A vector containing what groups are 'active'. Meaning calculations and graphing will use these groupings. If no value is supplied it defaults to all grouping variables. Can change active_group later using `active_group() <- 'myvalue'`. 


### Vector inputs

In the case of vector inputs, the vectors are cbinded to `data` if data is supplied. Sftrack returns an error if the column name already exists in the data.frame and `overwrite_names` remains FALSE. Default column names are: group = 'sft_group', time = 'sft_timestamp', error = 'sft_error'. Names can be overwitten using `overwrite_names = TRUE`. At present Geometry defaults as 'Geometry' and at present can be changed later via `sf` but not via `sftrack` as we make a new sf object when we make an sftrack object. 


```{r}
library("sftrack")

#data
data("raccoon", package = "sftrack")

#xyz
coords <- raccoon[,c("longitude", "latitude")]
crs <- 4326
#groupings
group <- list(id = raccoon$animal_id,month = as.POSIXlt(raccoon$timestamp)$mon+1)
active_group <- c("id","month")
#time
time <- as.POSIXct(raccoon$timestamp, tz = "EST")
#error
error <- raccoon$fix
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, 
                         active_group = active_group, time = time, 
                         crs = crs, error = error)

head(my_sftrack)

```

As you can see in this case the data is not overwritten, but extra columns added with the correct data.

-------------

### data.frame inputs  

Data.frame inputs generally describe the columns that represent the variables in `data`. If the columns are not found in `data`, an error is returned. 


```{r}
raccoon$time <- as.POSIXct(raccoon$timestamp, tz = "EST")
raccoon$month <- as.POSIXlt(raccoon$timestamp)$mon + 1

coords <- c("longitude", "latitude")
group <- c(id = "animal_id", month = "month")
time <- "time"
error <- "fix"

my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time, error = error)

head(my_sftraj)
```

## Conversion mode

`as_sftrack()` and `as_sftraj()` also accept other data types, but the arguments for each differ depending on the class its converting from. It currently accepts, `sf`, `ltraj` (from adehabitatLT), and eventually `tibbles` (tidyverse) and `track` (trajectories). 

### Import from ltraj

To read in an ltraj object all you need is an ltraj object created in `adehabitatLT`. All relevant information is taken from the object. The burst as defined in an ltraj is slightly different than sftracks groupings, so it assumes the ltraj 'burst' is the `id` field of the sftrack grouping object.

```{r, message = FALSE}
library("adehabitatLT")

ltraj_df <- as.ltraj(xy = raccoon[,c('longitude','latitude')], date = as.POSIXct(raccoon$timestamp),
 id = raccoon$animal_id, typeII = TRUE,
 infolocs = raccoon[,1:6])

my_sf <- as_sftrack(ltraj_df)
head(my_sf)

```

### sf objects

`sf` objects are handled similarly as data.frame inputs except you do not need to input any information about the coordinates or projection. Grouping and time are still required. The sf spatial object must be in an `sfc_POINT` column. 

```{r}
library("sf")
df1 <- raccoon[!is.na(raccoon$latitude),]
sf_df <- st_as_sf(df1, coords=c("longitude","latitude"), crs = crs)
group <- c(id = "animal_id")
time_col <- "time"

new_sftraj <- as_sftraj(sf_df, group = group, time = time_col) 
head(new_sftraj)

new_sftrack <- as_sftrack(sf_df, group = group, time = time_col) 
head(new_sftrack)

```

### Intra-package conversion between classes

Additionally `as_sftrack` and `as_sftraj` can convert back and forth between classes with no loss in information.

```{r}
# Make tracks from raw data
coords <- c("longitude","latitude")
group <- c(id = "animal_id", month = "month")
time <- "time"
error <- "fix"

my_sftraj <- as_sftraj(data = raccoon, coords = coords, group = group, time = time, error = error)
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error)

# Convert between types
new_sftrack <- as_sftrack(my_sftraj)
#head(new_sftrack)
new_sftraj <- as_sftraj(my_sftrack)
#head(new_sftraj)

identical(my_sftraj,new_sftraj)
identical(my_sftrack,new_sftrack)
```
### Duplicated data

A common issue with movement data is when duplicated gps time stamps are logged for a sensor. When this happens it can be impossible for `sftrack` to know which point to use. For this reason, sftrack returns an error if any `grouping` + `time` combinations are duplicated. 


```{r}

raccoon$time[1] <- raccoon$time[2]
try(as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error))

```
To help determine which rows are duplicated you can use the `which_duplicated` function to check your inputs. After which you can delete the superfluous rows and try again:

```{r}

which_duplicated(data = raccoon , group = group, time = time)
raccoon <- raccoon[-2,]
my_sftrack <- as_sftrack(data = raccoon, coords = coords, group = group, time = time, error = error)

```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_NA_coords}
\alias{check_NA_coords}
\title{Check if coordinates contain NAs in some columns but not others}
\usage{
check_NA_coords(xyz)
}
\arguments{
\item{xyz}{a data.frame of xy or xyz coordinates}
}
\description{
Check if coordinates contain NAs in some columns but not others
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/burst_tj.R
\name{active_group}
\alias{active_group}
\title{Access the active_group value}
\usage{
active_group(x)
}
\arguments{
\item{x}{a c_grouping}
}
\description{
The active group is the combination of group names to group the data sets.
The active_group acts essentially like a paste(names_of_groups, sep = '_') grouping variable.
}
\examples{
#'
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
mb1 <- make_c_grouping(x = burstz, active_group = c("id", "month"))

# see the current active burst
active_group(mb1)

# change the active burst
active_group(mb1) <- "id"

# Using a full data set
my_track <- as_sftrack(raccoon,
  time = "timestamp",
  error = NA, coords = c("longitude", "latitude"),
  group = burstz
)

summary(my_track)

# change active group
active_group(my_track$sft_group) <- "id"

summary(my_track)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_calculation.R
\name{make_step_geom}
\alias{make_step_geom}
\title{Calculate step geometries given a set of groupings, time, and geometries}
\usage{
make_step_geom(group, time_data, geometry)
}
\arguments{
\item{group}{a c_grouping object}

\item{time_data}{time vector}

\item{geometry}{the geometery data from either sf or sf_track. Must be an sf geometry class}
}
\description{
This calculates step geometries as individual line segments based on the active_group
}
\examples{
#'
library("sf")
geom <- st_as_sf(data.frame(
  x = c(1, 2, 2, 5),
  y = c(0, 1, 5, 7)
), coords = c("x", "y"))

burst <- list(id = rep(1, 4))
time <- 1:4

cg <- make_c_grouping(burst)

make_step_geom(
  group = cg,
  geometry = geom$geometry,
  time_data = time
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/burst_tj.R
\name{grouping-class}
\alias{grouping-class}
\alias{make_s_group}
\alias{make_c_grouping}
\alias{c.s_group}
\alias{c.c_grouping}
\title{A class to group movement data}
\usage{
make_s_group(x)

make_c_grouping(x = NULL, active_group = NULL)

\method{c}{s_group}(...)

\method{c}{c_grouping}(..., recursive = FALSE)
}
\arguments{
\item{x}{a list containing named grouping variables, one item must be named 'id'. ex: list(id = 1, month = 'may'). For a c_grouping: A list of s_groups or
a list of equal length named vectors which will be combined to create a c_grouping. ex: list(x = 1st_vector, y = 2nd_vector)}

\item{active_group}{a vector of the names of the groups to be considered 'active'.}

\item{...}{objects to be pasted together into a c_grouping}

\item{recursive}{ignored}
}
\description{
This class describes grouping variables for movement data.
The grouping object is composed of a list with named vectors.
One of which must be 'id', this is the id of subject being monitored (commonly animal id in movement data)
Can be any number of groups after that.
}
\details{
A grouping is a list of possible categories to group the data. The 'active group'
of these is the current grouping variables to be considered for analysis. The 'active group' can be any combination of the categories in a burst,
and can change with the use of `active_group()`.

An `s_group` is a single row group. It is a 1xn dimensional list with any length(n) > 1.
Atleast one of the groups must be named 'id' which is the subjects id.

A `c_grouping` is a collection of `s_groups`s, it is a data.frame with dimensions of 1xnrow(data).
One c_grouping has one 'active group' which describes the set of names in each s_group to group the data. When you change the 'active group', calculations and plots change accordingly
to the new grouping levels.

You can create bursts with make_s_group and make_c_grouping.
}
\examples{
# Make a single group
#'
make_s_group(x = list(id = "CJ11", month = 3, height = 10))

# Make a c_grouping
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
mb1 <- make_c_grouping(x = burstz, active_group = c("id", "month"))
str(mb1)

# Make a multi_burst from many ind_bursts
a <- make_s_group(list(id = 1, year = 2020))
b <- make_s_group(list(id = 1, year = 2020))
c <- make_s_group(list(id = 2, year = 2020))

c(a, b, c)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_time}
\alias{check_time}
\title{Check if time is integer or posix}
\usage{
check_time(time)
}
\arguments{
\item{time}{a vector of time}
}
\description{
Check if time is integer or posix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{fix_zero}
\alias{fix_zero}
\title{fix 0's to NAs in latitude and longitude}
\usage{
fix_zero(xyz)
}
\arguments{
\item{xyz}{a data.frame of xy or xyz coordinates}
}
\value{
returns a data.frame with 0s replaced with NAs
}
\description{
fix 0's to NAs in latitude and longitude
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_tj.R
\name{plot_sftrack}
\alias{plot_sftrack}
\alias{plot.sftrack}
\alias{plot.sftraj}
\title{Methods for plotting sftrack/sftraj}
\usage{
\method{plot}{sftrack}(x, y, key.pos, key.width, ...)

\method{plot}{sftraj}(x, y, key.pos, key.width, ..., step_mode)
}
\arguments{
\item{x}{'sftrack' or 'sftraj' object}

\item{y}{ignored}

\item{key.pos}{Integer; side to plot a color key: 1 bottom, 2 left,
3 top, 4 right; set to NULL to omit key, or -1 to select
automatically (defaults to 4; see \code{\link[sf]{plot_sf}} for
more details).}

\item{key.width}{Amount of space reserved for the key, including
labels (see \code{\link[sf]{plot_sf}} for more details).)}

\item{...}{Further arguments passed to 'plot.sf'. Among others,
arguments for the key are set differently in 'sftrack' to allow
for longer labels by default (but can be nevertheless
adjusted).}

\item{step_mode}{Logical; whether to plot in step mode, see
details, defaults to TRUE, unless there are more than 10,000
steps.}
}
\description{
Methods for plotting sftrack/sftraj

Methods for plotting sftrack/sftraj
}
\details{
Step mode refers to considering the trajectory as
    individual 'steps', in the case of plot this means it will plot
    each line & point individually. This approach is much slower to
    plot with large objects, and is thus turned off when
    n(steps)>10,000.  The alternative, much faster method is to
    merge the steps into a multilinestring as continuous lines.
}
\examples{

## Prepare an 'sftrack' object:
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- c(id = "animal_id")
my_sftrack <- as_sftrack(raccoon,
  time = "timestamp",
  coords = c("longitude", "latitude"),
  group = burstz
)

## Plotting with sftrack is just like sf. `...` will accept most
## arguments as 'plot.sf':
plot(my_sftrack, axes = TRUE, lwd = 5, cex = 5, bgc = "gray50")

## sftraj will as well for the most part; however it is a more complex
## structure that combines points and steps (in step mode):
my_sftraj <- as_sftraj(raccoon,
  time = "timestamp",
  coords = c("longitude", "latitude"),
  group = burstz
)
plot(my_sftraj, lwd = 5, cex = 5, bgc = "gray50", graticule = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/burst_tj.R
\name{calc_sort_index}
\alias{calc_sort_index}
\title{Calculate a new sort index for groups}
\usage{
calc_sort_index(x, active_group = NA)
}
\arguments{
\item{x}{group or sftrack object}

\item{active_group}{(optional), a new active group. If not included, defaults  to the active group (if a c_grouping) or the group names}
}
\description{
Calculate a new sort index for groups
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_util.R
\name{which_duplicated}
\alias{which_duplicated}
\title{Which grouping/time stamp combos are duplicated.}
\usage{
which_duplicated(data = data.frame(), group, time)
}
\arguments{
\item{data}{a data.frame containing burst or time data (if necessary)}

\item{group}{a list where each entry is a vector of groupings where length == nrow(data)|nrow(time). Or a character vector describing the column name they are located in data}

\item{time}{a vector of as.POSIXct time, or a character of the column name where it can be found in data}
}
\description{
This function returns a data.frame of which rows are duplicated and their time stamps.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_calculation.R
\name{step_recalc}
\alias{step_recalc}
\title{recalculate step geometry}
\usage{
step_recalc(x, return = FALSE)
}
\arguments{
\item{x}{an sftraj object.}

\item{return}{return step_geometry instead of replacing sftraj object with new step geometry. Defaults to FALSE}
}
\description{
Step geometeries in sftraj objects are linestrings going from t1 to t2 of a 'step'. As these are stored at the row level they are not dynamic to changes in t2.
step_recalc allows you to recalculate these geometeries if your data.frame has changed because of subsetting or filtering.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sftrack.R
\name{new_sftrack}
\alias{new_sftrack}
\title{Define an sftrack}
\usage{
new_sftrack(data, group_col, sf_col, time_col, error_col = NA)
}
\arguments{
\item{data}{data.frame with multi_burst column, geometry column, time_col (integer/POSIXct), and error column (optional)}

\item{group_col}{column name of grouping info in `data`}

\item{sf_col}{column name of geometry info in `data`}

\item{time_col}{column name of time info in `data`}

\item{error_col}{column name of error info in `data`}
}
\description{
Define an sftrack
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sftrack.R
\name{as_sftrack}
\alias{as_sftrack}
\alias{as_sftrack.data.frame}
\alias{as_sftrack.sftraj}
\alias{as_sftrack.ltraj}
\alias{as_sftrack.sf}
\title{Convert objects into sftrack objects.}
\usage{
as_sftrack(data = data.frame(), ...)

\method{as_sftrack}{data.frame}(
  data = data.frame(),
  ...,
  coords = c("x", "y"),
  group = "id",
  active_group = NA,
  time = "time",
  error = NA,
  crs = NA,
  zeroNA = FALSE,
  group_name = "sft_group",
  timestamp_name = "sft_timestamp",
  error_name = "sft_error",
  overwrite_names = FALSE
)

\method{as_sftrack}{sftraj}(data, ...)

\method{as_sftrack}{ltraj}(data, ...)

\method{as_sftrack}{sf}(
  data,
  ...,
  coords,
  group,
  active_group = NA,
  time,
  error = NA,
  group_name = "sft_group",
  timestamp_name = "sft_timestamp",
  error_name = "sft_error",
  overwrite_names = FALSE
)
}
\arguments{
\item{data}{a data.frame of the movement data, if supplied all data.frame inputs, than is optional}

\item{...}{extra information to be passed on to as_sftrack}

\item{coords}{a character vector describing where the x,y,z coordinates are located in `data` or a list with x,y,z (optional) vectors}

\item{group}{a list of named vectors describing multiple grouping variables or  a character vector naming the other grouping columns in `data`.}

\item{active_group}{a character vector of the burst names to be 'active' to group data by for analysis}

\item{time}{a vector of time information, can be either POSIX or an integer or a character string naming the column in `data` where the time information is located}

\item{error}{(optional) a vector of error information for the movement dataa character string naming the column in `data` where the error information is located}

\item{crs}{Coordinate reference system to be assigned; object of class `crs`. Defaults to NA}

\item{zeroNA}{logical whether to convert 0s in spatial data into NAs. Defaults to FALSE.}

\item{group_name}{(optional) new column name for grouping data}

\item{timestamp_name}{(optional) new column name for time data}

\item{error_name}{(optional) new column name for error data}

\item{overwrite_names}{T/F Whether to overwrite data if a group/time/error column name is supplied but already in data}
}
\description{
This function converts x,y,z data into an sftrack object with a sf_geometry column of sf_POINTS.
Creates a `grouping` column to group movement data and sets dedicated time and error columns.

Raw data inputted in two ways: vector or data.frame. 'Vector' inputs gives the argument as a vector where
length = nrow(data). 'Data.frame' inputs gives the arguments as the column name of `data` where the input can be found.
Either input is allowed on any given argument.

Some options are global and required regardless
}
\details{
Convert objects into sftrack objects.
}
\examples{
#'
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
# Input is a data.frame
my_track <- as_sftrack(raccoon,
  group = burstz, time = "timestamp",
  error = NA, coords = c("longitude", "latitude")
)

# Input is a ltraj
library("adehabitatLT")
ltraj_df <- as.ltraj(
  xy = raccoon[, c("longitude", "latitude")],
  date = as.POSIXct(raccoon$timestamp),
  id = raccoon$animal_id, typeII = TRUE,
  infolocs = raccoon[, 1:6]
)

my_sftrack <- as_sftrack(ltraj_df)
head(my_sftrack)

# Input is a sf object
library("sf")
df1 <- raccoon[!is.na(raccoon$latitude), ]
sf_df <- st_as_sf(df1, coords = c("longitude", "latitude"))

new_sftrack <- as_sftrack(sf_df, group = c(id = "animal_id"), time = "timestamp")
head(new_sftrack)

# Input is an sftraj object
my_traj <- as_sftraj(raccoon,
  time = "timestamp",
  error = NA, coords = c("longitude", "latitude"),
  group = burstz
)

new_track <- as_sftrack(my_traj)
head(new_track)
######################
# Builder
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_util.R
\name{is_linestring}
\alias{is_linestring}
\title{Is a trajectory geometry a linestring or a point}
\usage{
is_linestring(x)
}
\arguments{
\item{x}{an sftraj object}
}
\description{
A step is a movement from one point to the next, with an sftraj object
this manifests as a linestring. If, however, one of these two points is missing, the sftraj
is created as a geometery collection of two points, the beginning and the end point, where one
of the steps is NA. This function checks a trajectory geometry if its a linestring and returns
a vector of T/F.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_util.R
\name{merge_traj}
\alias{merge_traj}
\title{Merge connected lines and create an sf object}
\usage{
merge_traj(x)
}
\arguments{
\item{x}{an sftraj object}
}
\description{
This function returns a sf object grouped by each burst with a geometry column of multilinestrings for each grouping
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/burst_tj.R
\name{active_group<-}
\alias{active_group<-}
\alias{active_group_replace}
\alias{active_group<-.sftrack}
\alias{active_group<-.sftraj}
\alias{active_group<-.c_grouping}
\title{Set new active group}
\usage{
active_group(x) <- value

\method{active_group}{sftrack}(x) <- value

\method{active_group}{sftraj}(x) <- value

\method{active_group}{c_grouping}(x) <- value
}
\arguments{
\item{x}{sftrack/sftraj/c_grouping/s_group object}

\item{value}{character vector of the grouping names to make active}
}
\description{
Set new active group
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/burst_tj.R
\name{group_labels}
\alias{group_labels}
\alias{group_labels.sftrack}
\alias{group_labels.sftraj}
\alias{group_labels.c_grouping}
\title{Shows grouping labels created from the s_group and the c_grouping}
\usage{
group_labels(x)

\method{group_labels}{sftrack}(x)

\method{group_labels}{sftraj}(x)

\method{group_labels}{c_grouping}(x)
}
\arguments{
\item{x}{a sftrack or grouping object}
}
\description{
Shows grouping labels created from the s_group and the c_grouping
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_t_regular}
\alias{check_t_regular}
\title{Check if time is regular for each burst and returns logical for each burst}
\usage{
check_t_regular(x)
}
\arguments{
\item{x}{an sftrack/sftraj object}
}
\description{
Check if time is regular for each burst and returns logical for each burst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raccoon_dataset.R
\docType{data}
\name{raccoon}
\alias{raccoon}
\title{Movements of two raccoons in an urban park in Florida}
\format{
A data frame with 445 rows and 10 variables:
\describe{
  \item{animal_id}{ID of individual. TTP: tree tops park, i.e the tagging site.}
  \item{timestamp}{The date and time of gps fix in UTC}
  \item{latitude}{Latitude in degrees}
  \item{longitude}{Longitude in degrees}
  \item{height}{Altitude in meters based on satellite positios}
  \item{hdop}{Horizontal precision}
  \item{vdop}{Vertical precision}
  \item{fix}{The number of satellite fixes}
}
}
\usage{
raccoon
}
\description{
A dataset of two raccoons collared with GPS collars for one month in January 2019
in Tree Tops Park, Broward County, Florida, US.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_util.R
\name{summary_sftrack}
\alias{summary_sftrack}
\title{Summarize sftrack objects}
\usage{
summary_sftrack(x)
}
\arguments{
\item{x}{an sftrack object}
}
\description{
Summarize sftrack objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{dup_timestamp}
\alias{dup_timestamp}
\title{check that time is unique}
\usage{
dup_timestamp(x, time)
}
\arguments{
\item{x}{An sftrack/sftraj object or a multi_burst}

\item{time}{vector of time, not required if given a sftrack object.}
}
\description{
check that time is unique
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_tj.R
\name{geom_sftrack}
\alias{geom_sftrack}
\alias{geom_sftrack.sftrack}
\alias{geom_sftrack.sftraj}
\title{Function to plot sftrack objects in ggplot}
\usage{
geom_sftrack(mapping, data, ...)

\method{geom_sftrack}{sftrack}(mapping = ggplot2::aes(), data = NULL, ...)

\method{geom_sftrack}{sftraj}(mapping = ggplot2::aes(), data = NULL, ..., step_mode = FALSE)
}
\arguments{
\item{mapping}{mapping aesthetics for ggplot.}

\item{data}{the sftraj or sftrack object.}

\item{...}{arguments to passed to ggplot}

\item{step_mode}{TRUE/FALSE, whether to plot in step_mode, See details}
}
\description{
This function can be added to ggplot() to plot an sftrack and sftraj
Function does not yet work with ggplot grammer so you must but data= in this function
}
\details{
step mode refers to considering the trajectory as individual 'steps', in the case of plot this means it will
plot each line & point individually. This approach is much slower to plot when n(steps)>10,000.
The alternative method is to merge the steps into a multilinestring of continuous lines. This is much faster to plot.
}
\examples{
#'
require("ggplot2")
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- c(id = "animal_id")

# sftraj will as well for the most part, however as its a more complex
# structure to speed up plotting.
my_sftraj <- as_sftraj(raccoon,
  time = "timestamp",
  coords = c("longitude", "latitude"),
  group = burstz
)

ggplot() +
  geom_sftrack(data = my_sftraj)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_names_exist}
\alias{check_names_exist}
\title{Check if a set of column names are found in a data frame and return an error if not}
\usage{
check_names_exist(data, names)
}
\arguments{
\item{data}{a data.frame to check names against}

\item{names}{the inputted column names}
}
\description{
Check if a set of column names are found in a data frame and return an error if not
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_group_id}
\alias{check_group_id}
\title{Check there is a grouping id present}
\usage{
check_group_id(x)
}
\arguments{
\item{x}{a c_grouping}
}
\description{
Check there is a grouping id present
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/burst_tj.R
\name{group_names}
\alias{group_names}
\alias{group_name}
\alias{group_names.c_grouping}
\alias{group_names.sftraj}
\alias{group_names.sftrack}
\title{Display the levels of the sort index}
\usage{
group_names(x)

\method{group_names}{c_grouping}(x)

\method{group_names}{sftraj}(x)

\method{group_names}{sftrack}(x)
}
\arguments{
\item{x}{sftrack/sftraj/c_grouping/s_group object}
}
\description{
Display the levels of the sort index
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sftrack.R
\name{Print_sftrack_objects}
\alias{Print_sftrack_objects}
\alias{print.sftrack}
\title{Print methods for sftrack}
\usage{
\method{print}{sftrack}(x, n_row, n_col, ...)
}
\arguments{
\item{x}{sftraj object}

\item{n_row}{Integer of number of rows to display. Defaults to global option default if non supplied}

\item{n_col}{Integer of number of columns to display + required sftrack columns (burst, geometry, time, and error). Defaults to global option default if non supplied}

\item{...}{other arguments passed onto print}
}
\description{
Print methods for sftrack
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_group_names}
\alias{check_group_names}
\title{Are group names equivalent for each s_group?}
\usage{
check_group_names(x)
}
\arguments{
\item{x}{a c_grouping}
}
\description{
Are group names equivalent for each s_group?
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_util.R
\name{traj_geom}
\alias{traj_geom}
\alias{pts_traj}
\alias{coord_traj}
\title{Return a list of sf_POINTS or a data.frame from a sftraj object}
\usage{
pts_traj(traj, sfc = FALSE)

coord_traj(traj)
}
\arguments{
\item{traj}{a trajectory geometery from sf_traj}

\item{sfc}{TRUE/FALSE should the return by an sfc or a list of points. Defaults to FALSE}
}
\description{
Return a list of sf_POINTS or a data.frame from a sftraj object
}
\examples{
#'
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
# Input is a data.frame
my_traj <- as_sftraj(raccoon,
  time = "timestamp",
  error = NA, coords = c("longitude", "latitude"),
  group = burstz
)
print(my_traj, 5, 10)

# extract a list of points
pts_traj(my_traj)[1:10]

# or a data.frame of points
coord_traj(my_traj)[1:10]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_ordered}
\alias{check_ordered}
\title{Checks if grouping is ordered by time and then outputs the correct order}
\usage{
check_ordered(group, time_data, return = TRUE)
}
\arguments{
\item{group}{a c_grouping}

\item{time_data}{a vector of time}

\item{return}{T/F return the new order or just run check?}
}
\description{
Checks if grouping is ordered by time and then outputs the correct order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_calculation.R
\name{step_metrics}
\alias{step_metrics}
\title{Calculates step metrics including distance, dt, dx, and dy.}
\usage{
step_metrics(sftraj)
}
\arguments{
\item{sftraj}{an sftrack/sftraj object. sftrack objects will be converted to sftraj internally for calculation.}
}
\description{
Calculates step metrics including distance, dt, dx, and dy.
}
\examples{
#'
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
# Input is a data.frame
my_sftraj <- as_sftraj(raccoon,
  group = burstz, time = "timestamp",
  error = NA, coords = c("longitude", "latitude")
)

step_metrics(my_sftraj)[1:10, ]
}
